# Data from: Pyykko, P. and Atsumi, M., Chem. Eur. J. 2009, 15, 186.
element_radii=[
    ["None",None],['H', 32],['He', 46],[ 'Li', 133],[ 'Be', 102],[ 'B', 85],[ 'C', 75],
    ['N', 71],['O', 63],['F', 64],[ 'Ne', 67],[ 'Na', 155],[ 'Mg', 139],
    ['Al', 126],['Si', 116],['P', 111],[ 'S', 103],[ 'Cl', 99],[ 'Ar', 96],
    ['K', 196],['Ca', 171],['Sc', 148],[ 'Ti', 136],[ 'V', 134],[ 'Cr', 122],
    ['Mn', 119],['Fe', 116],['Co', 111],[ 'Ni', 110],[ 'Cu', 112],[ 'Zn', 118],
    ['Ga', 124],['Ge', 121],['As', 121],[ 'Se', 116],[ 'Br', 114],[ 'Kr', 117],
    ['Rb', 210],['Sr', 185],['Y', 163],[ 'Zr', 154],[ 'Nb', 147],[ 'Mo', 138],
    ['Tc', 128],['Ru', 125],['Rh', 125],[ 'Pd', 120],[ 'Ag', 128],[ 'Cd', 136],
    ['In', 142],['Sn', 140],['Sb', 140],[ 'Te', 136],[ 'I', 133],[ 'Xe', 131],
    ['Cs', 232],['Ba', 196],['La', 180],[ 'Ce', 163],[ 'Pr', 176],[ 'Nd', 174],
    ['Pm', 173],['Sm', 172],['Eu', 168],[ 'Gd', 169],[ 'Tb', 168],[ 'Dy', 167],
    ['Ho', 166],['Er', 165],['Tm', 164],[ 'Yb', 170],[ 'Lu', 162],[ 'Hf', 152],
    ['Ta', 146],['W', 137],['Re', 131],[ 'Os', 129],[ 'Ir', 122],[ 'Pt', 123],
    ['Au', 124],['Hg', 133],[ 'Tl', 144],[ 'Pb', 144],[ 'Bi', 151],[ 'Po', 145],
    ['At', 147],['Rn', 142],[ 'Fr', 223],[ 'Ra', 201],[ 'Ac', 186],[ 'Th', 175],
    ['Pa', 169],['U', 170],[ 'Np', 171],[ 'Pu', 172],[ 'Am', 166],[ 'Cm', 166],
    ['Bk', 168],['Cf', 168],[ 'Es', 165],[ 'Fm', 167],[ 'Md', 173],[ 'No', 176],
    ['Lr', 161],['Rf', 157],[ 'Db', 149],[ 'Sg', 143],[ 'Bh', 141],[ 'Hs', 134],
    ['Mt', 129],['Ds', 128],['Rg', 121],[ 'Cn', 122],[ 'Nh', 136],[ 'Fl', 143],
    ['Mc', 162],['Lv', 175],['Ts', 165],['Og', 157]]
elements = tuple(i[0] for i in element_radii)
keywords = \
    ['1-bromo-2-methylpropane', '1-bromooctane', '1-bromopentane', '1-bromopropane', '1-butanol',
    '1-chlorohexane', '1-chloropentane', '1-chloropropane', '1-decanol', '1-fluorooctane', '1-heptanol',
    '1-hexanol', '1-hexene', '1-hexyne', '1-iodobutane', '1-iodohexadecane', '1-iodopentane',
    '1-iodopropane', '1-nitropropane', '1-nonanol', '1-pentanol', '1-pentene', '1-propanol',
    '1-trichloroethane', '2-bromopropane', '2-butanol', '2-chlorobutane', '2-dibromoethane',
    '2-dichloroethene', '2-dimethylcyclohexane', '2-ethanediol', '2-heptanone', '2-hexanone',
    '2-methoxyethanol', '2-methyl-1-propanol', '2-methyl-2-propanol', '2-methylpentane',
    '2-methylpyridine', '2-nitropropane', '2-octanone', '2-pentanone', '2-propanol', '2-propen-1-ol',
    '2-trichloroethane', '2-trifluoroethanol', '3-methylpyridine', '3-pentanone', '4-dimethylpentane',
    '4-dimethylpyridine', '4-dioxane', '4-heptanone', '4-methyl-2-pentanone', '4-methylpyridine',
    '4-trimethylbenzene', '4-trimethylpentane', '5-nonanone', '6-dimethylpyridine', 'a-chlorotoluene',
    'aceticacid', 'acetone', 'acetonitrile', 'acetophenone', 'allcheck', 'aniline', 'anisole', 'apfd',
    'argon', 'b1b95', 'b1lyp', 'b3lyp', 'b3p86', 'b3pw91', 'b971', 'b972', 'b97d', 'b97d3', 'benzaldehyde',
    'benzene', 'benzonitrile', 'benzylalcohol', 'betanatural', 'bhandh', 'bhandhlyp', 'bromobenzene',
    'bromoethane', 'bromoform', 'butanal', 'butanoicacid', 'butanone', 'butanonitrile', 'butylamine',
    'butylethanoate', 'calcall', 'calcfc', 'cam-b3lyp', 'carbondisulfide', 'carbontetrachloride',
    'cartesian', 'checkpoint', 'chkbasis', 'chlorobenzene', 'chloroform', 'cis-1', 'cis-decalin',
    'connectivity', 'counterpoise', 'cyclohexane', 'cyclohexanone', 'cyclopentane', 'cyclopentanol',
    'cyclopentanone', 'd95v', 'decalin-mixture', 'def2qzv', 'def2qzvp', 'def2qzvpp', 'def2sv', 'def2svp',
    'def2svpp', 'def2tzv', 'def2tzvp', 'def2tzvpp', 'density', 'densityfit', 'dibromomethane',
    'dibutylether', 'dichloroethane', 'dichloromethane', 'diethylamine', 'diethylether', 'diethylsulfide',
    'diiodomethane', 'diisopropylether', 'dimethyldisulfide', 'dimethylsulfoxide', 'diphenylether',
    'dipropylamine', 'e-2-pentene', 'empiricaldispersion', 'ethanethiol', 'ethanol', 'ethylbenzene',
    'ethylethanoate', 'ethylmethanoate', 'ethylphenylether', 'extrabasis', 'extradensitybasis', 'finegrid',
    'fluorobenzene', 'formamide', 'formicacid', 'freq', 'full', 'gd3bj', 'genecp', 'geom', 'gfinput',
    'gfprint', 'hcth', 'hcth147', 'hcth407', 'hcth93', 'heptane', 'hexanoicacid', 'hissbpbe', 'hseh1pbe',
    'integral', 'iodobenzene', 'iodoethane', 'iodomethane', 'isopropylbenzene', 'isoquinoline', 'kcis',
    'krypton', 'lanl2dz', 'lanl2mb', 'lc-wpbe', 'loose', 'm-cresol', 'm-xylene', 'm062x', 'm06hf', 'm06l',
    'm11l', 'maxcycles', 'maxstep', 'mesitylene', 'methanol', 'methylbenzoate', 'methylbutanoate',
    'methylcyclohexane', 'methylethanoate', 'methylmethanoate', 'methylpropanoate', 'minimal', 'mn12l',
    'mn12sx', 'modredundant', 'mpw1lyp', 'mpw1pbe', 'mpw1pw91', 'mpw3pbe', 'n-butylbenzene', 'n-decane',
    'n-dimethylacetamide', 'n-dimethylformamide', 'n-dodecane', 'n-hexadecane', 'n-hexane',
    'n-methylaniline', 'n-methylformamide-mixture', 'n-nonane', 'n-octane', 'n-octanol', 'n-pentadecane',
    'n-pentane', 'n-undecane', 'n12sx', 'nitrobenzene', 'nitroethane', 'nitromethane', 'noeigentest',
    'nofreeze', 'noraman', 'nosymm', 'nprocshared', 'o-chlorotoluene', 'o-cresol', 'o-dichlorobenzene',
    'o-nitrotoluene', 'o-xylene', 'o3lyp', 'ohse1pbe', 'ohse2pbe', 'oniom', 'output', 'p-isopropyltoluene',
    'p-xylene', 'pbe1pbe', 'pbeh', 'pbeh1pbe', 'pentanal', 'pentanoicacid', 'pentylamine',
    'pentylethanoate', 'perfluorobenzene', 'pkzb', 'population', 'propanal', 'propanoicacid',
    'propanonitrile', 'propylamine', 'propylethanoate', 'pseudo', 'pw91', 'pyridine', 'qst2', 'qst3',
    'quinoline', 'qzvp', 'rdopt', 'read', 'readfc', 'readfreeze', 'readopt', 'readoptimize', 'regular',
    'restart', 's-dioxide', 'savemixed', 'savemulliken', 'savenbos', 'savenlmos', 'scrf', 'sddall',
    'sec-butylbenzene', 'sogga11', 'sogga11x', 'solvent', 'spinnatural', 'tert-butylbenzene',
    'tetrachloroethene', 'tetrahydrofuran', 'tetrahydrothiophene-s', 'tetralin', 'thcth', 'thcthhyb',
    'thiophene', 'thiophenol', 'tight', 'toluene', 'tpss', 'tpssh', 'trans-decalin', 'tributylphosphate',
    'trichloroethene', 'triethylamine', 'tzvp', 'ultrafine', 'uncharged', 'v5lyp', 'verytight', 'vp86',
    'vsxc', 'vwn5', 'water', 'wb97', 'wb97x', 'wb97xd', 'wpbeh', 'x3lyp', 'xalpha', 'xenon',
    'xylene-mixture']
subshells = ("1s", "2s", "2p", "3s", "3p", "4s", "3d", "4p", "5s", "4d", "5p",
             "6s", "4f", "5d", "6p", "7s", "5f", "6d", "7p", "8s", "5g", "6f",
             "7d", "8p", "9s")
subshell_size = (2,2,6,2,6,2,10,6,2,10,6,2,14,10,6,2,14,10,6,2,18,14,10,6,2)



import os, shutil
from raachem.file_class.log import LogFile
from raachem.file_class.gjf import GjfFile
from raachem.util.gen_purp import file_weeder, read_item, preferences, timeit, is_str_float


def e_analysis(weeded_list):
	"""Analyses Gaussian log files in weeded_list and prints out a txt file with results in current directory"""
	file_name = "Extracted_data.txt"
	out = []
	out.append("{} File(s) were analyzed in total.".format(len(weeded_list)))
	out.append("{:^29}|{:^20}|{:^20}|{:^20}".format("Name", "Sum of Elet an TC", "Thermal Correct", "last SCF"))
	rel_e = []
	for i in weeded_list:
		log = LogFile(read_item(i))
		t = log.thermal
		try: last_SCF = log.scf_done[-1][-1]
		except: last_SCF = "None"
		if t == False: out.append("{:<30} no_thermo_data_found_Last_SCF: {:>20}".format(i, last_SCF)); continue
		out.append("{:<30}{:>20}{:>20}{:>20}".format(str(i), str(t[7]), str(t[3]), str(last_SCF)))
		rel_e.append([i, float(t[7]) * 627.5095, float(last_SCF) * 627.5095, float(t[3]) * 627.5095])
	if len(rel_e) > 1:
		print("Do you want the free energies to be reported relative to which item?")
		print("Enter 0 if you don't want to analyze them)")
		for idx, entry in enumerate(rel_e):
			print(" {:<4}{:<20}{:>25}".format(idx + 1, entry[0], round(entry[1], 2)))
		while True:
			option = input()
			if option in [str(a) for a in range(len(rel_e)+1)]: option = int(option); break
			else: print("Could notunderstand request")
		if option == 0: print("Leaving analysis\n")
		else:
			out.append("\nFree energies relative to {}, (Name,G,H,TC):".format(rel_e[option-1][0]))
			for i in rel_e:
				a = [i[0], *[round(i[n] - rel_e[option-1][n], 2) for n in [1,2,3]]]
				out.append("{:<26}{:>26}{:>10}{:>10}".format(*a))
			out.append("\n")
	with open(file_name,mode="w",newline="\n") as file: file.write("\n".join(out))
	print("\nDone! \nPlease lookup:\n\n" + os.path.join(os.getcwd(), file_name), "\n")

def rel_scf(list=False):
	energies = []
	for i in file_weeder([".log"]):
		log = LogFile(read_item(i))
		energies.append([log.name, log.scf_done[-1][1], log.normal_termin])
	if len(energies) == 0: print("No .log files in current folder"); return
	energies = [[i[0], float(i[1]), i[2]] for i in energies]
	energies.sort(key=lambda x: x[1])
	min_e = energies[0][1]
	energies = [[i[0], (i[1] - min_e) * 627.509, i[2]] for i in energies]
	if list == False: print("\n".join(["{:>30}{:>15f}{:>5}".format(*l) for l in energies]))
	elif list == True: return energies

@timeit
def csv_e_analysis():
	def evaluate_list(folder, files=set()):
		for file in os.listdir(folder):
			if os.path.isdir(os.path.join(folder, file)):
				evaluate_list(os.path.join(folder, file), files)
			elif any([file.endswith(".log"),file.endswith(".xyz"),file.endswith(preferences.gauss_ext)]):
				files.add(os.path.join(folder, os.path.splitext(file)[0]))
		return files
	def evaluate_file(a, i,last):
		row = {a[1]:"-" for a in preferences.e_options}
		try:
			#BLANK
			row.update({"Blank": "-"})
			row.update({"Eh to kcal/mol":"627.5095"})
			row.update({"Eh to kJ/mol": "2625.5002"})

			#FILE PROPERTIES
			row.update({"Filename":os.path.basename(a)})

			#FOLDER PROPERTIES
			fold_name = os.path.dirname(a)
			row.update({"rel_path":os.path.relpath(fold_name, os.getcwd())})
			row.update({"Folder": 'HYPERLINK("{}";"{}")'.format(fold_name,os.path.relpath(fold_name, os.getcwd()))})

			#XYZ PROPERTIES
			xyz_name = a + ".xyz"
			is_xyz = os.path.isfile(os.path.relpath(xyz_name, os.getcwd()))
			row.update({".xyz":'HYPERLINK("{}";"Link")'.format(xyz_name) if is_xyz else "-"})

			#INPUT PROPERTIES
			inp_name = a + preferences.gauss_ext
			is_inp = os.path.isfile(os.path.relpath(inp_name, os.getcwd()))
			inp = GjfFile(read_item(os.path.relpath(inp_name, os.getcwd()))) if is_inp else False
			if inp:
				row.update({preferences.gauss_ext:'HYPERLINK("{}";"Link")'.format(inp_name) if is_inp else "-"})
				row.update({preferences.gauss_ext + "_#": inp.route_text() if inp else "-"})

			#LOG PROPERTIES
			log_name = a + ".log"
			is_log = os.path.isfile(os.path.relpath(log_name, os.getcwd()))
			log = LogFile(read_item(os.path.relpath(log_name, os.getcwd()))) if is_log else False
			if log:
				row.update({".log": 'HYPERLINK("{}";"Link")'.format(log_name) if log else "-"})
				row.update({".log_#": log.raw_route if log.raw_route else "-"})
				row.update({"E0":str(log.scf_done[-1][-1]) if log.scf_done else "-"})
				row.update({"iFreq": log.last_freq.n_ifreq() if log.last_freq else "-"})
				row.update({"E_ZPE": str(log.thermal[4]) if log.thermal else "-"})
				row.update({"E_tot": str(log.thermal[5]) if log.thermal else "-"})
				row.update({"H_corr": str(log.thermal[6]) if log.thermal else "-"})
				row.update({"G_corr": str(log.thermal[7]) if log.thermal else "-"})
				row.update({"E0+E_ZPE": str(log.thermal[0]) if log.thermal else "-"})
				row.update({"E0+E_tot": str(log.thermal[1]) if log.thermal else "-"})
				row.update({"E0+H_corr": str(log.thermal[2]) if log.thermal else "-"})
				row.update({"E0+G_corr": str(log.thermal[3]) if log.thermal else "-"})
				row.update({"Done?":"Yes" if log.normal_termin else "No"})
				row.update({"Error":log.error_msg if log else "-"})
				row.update({"HOMO": log.homo if log.homo else "-"})
				row.update({"LUMO": log.lumo if log.homo else "-"})
				row.update({"HOMO-LUMO": log.homo_lumo_gap if log.homo else "-"})
				row.update({"Charge": log.charge_mult[0] if log.charge_mult else "-"})
				row.update({"Mult": log.charge_mult[1] if log.charge_mult else "-"})
				row.update({"n_SCF":len(log.scf_done) if log.scf_done else "-"})
				row.update({"n_atoms": log.n_atoms if log.n_atoms else "-"})
				row.update({"TYP": log.calc_type if log.calc_type else "-"})
				row.update({"Needs refinement?": log.needs_ref()})
				row.update({"S**2 BA": log.s_squared[-1][1] if log.s_squared else "-"})
				row.update({"S**2 After": log.s_squared[-1][2] if log.s_squared else "-"})
				row.update({"LG": "\n".join(log.last_xyz_obj().return_print() if log.last_xyz_obj() else "-")})
				row.update({"LG": "-"})
				row.update({"MulkSpinDens": log.last_muliken_spin_density})
				row.update({"LastIntCoord": log.last_internal_coord})
				row.update({"MulkCharges": log.last_muliken_charges})
				row.update({"ESPCharges":log.last_chelpg_charges})
				row.update({"POPAnalysis":log.pop_analysis})
				row.update({"NPAAnalysis": log.npa_analysis})

		except Exception as e:
			print()
			print("\nError on file:\n{}\n".format(a));
			print(e, "\n")
		finally:
			if i+1 < last: print("\rEvaluating... {}/{}".format(i+1,last),end=" ")
			else: print("\rEvaluation done ({}/{}), saving '.xls' file...".format(i+1,last))
			return row
	aux_files_needed = ["LG","MulkSpinDens","LastIntCoord","MulkCharges","ESPCharges","POPAnalysis","NPAAnalysis"]
	folder_name = os.path.join(os.getcwd(), "linked_analysis_suplementary_files")
	if os.path.exists(os.path.join(folder_name)):
		if any(a in (preferences.e_options[n][1] for n in preferences.e_analysis_opt) for a in aux_files_needed):
			print("The following directory will be OVERWRITTEN!")
			print(folder_name)
			print("Are you sure? (y/n)")
			while True:
				if input() == "y": break
				else: print("Canceling...");return

	files = evaluate_list(os.getcwd())
	csv_list = [evaluate_file(file,i,len(files)) for i,file in enumerate(files)]
	if not csv_list: print("No .log files in {} directory".format(os.getcwd())); return
	csv_list.sort(key=lambda x: x["Folder"])
	try:
		import xlwt
		from xlwt import Workbook
	except ImportError:
		print("xlwt module is needed")
		return
	wb = Workbook()
	sheet1 = wb.add_sheet('Data')
	sheet2 = wb.add_sheet('Labels')
	style0 = xlwt.easyxf("", "#.0000000")

	for i_b,b in enumerate(preferences.e_options[n] for n in preferences.e_analysis_opt):
		sheet2.write(i_b, 0, b[1])
		sheet2.write(i_b, 1, b[0])
	for i_b,b in enumerate(preferences.e_options[n][1] for n in preferences.e_analysis_opt):
		sheet1.write(0, i_b, b)
	for i_a,a in enumerate(csv_list,start=1):
		for i_b,b in enumerate(preferences.e_options[n][1] for n in preferences.e_analysis_opt):
			if b in ("E0+G_corr", "E0", "E0+H_corr"):
				sheet1.write(i_a, i_b, float(a[b]) if is_str_float(a[b]) else a[b],style0)
			elif b in ("Folder",".xyz",preferences.gauss_ext,".log"):
				try: sheet1.write(i_a, i_b, xlwt.Formula(a[b]))
				except Exception: sheet1.write(i_a, i_b, a[b])
			elif b in aux_files_needed:
				if a[b] and a[b] != "-":
					filename = os.path.join(folder_name,a["rel_path"],a["Filename"]+"_"+b+".txt")
					try:
						os.makedirs(os.path.dirname(filename), exist_ok=True)
						if type(a[b]) == str:
								with open(filename, "w") as f:
									f.write(a[b])
								sheet1.write(i_a, i_b, xlwt.Formula('HYPERLINK("{}";"Link")'.format(filename)))
					except FileExistsError:
						print("Error while creating the following file:")
						print(filename)
						print("File already exists!")
				else:
					sheet1.write(i_a, i_b, "-")
			else: sheet1.write(i_a, i_b, a[b])
	try:
		wb.save("linked_analysis.xls")
	except PermissionError:
		print("Error while saving file!\nIs the file '{}' already open?".format("linked_analysis.xls"))

def deduplicate():
	print("Analyzing energies...")
	energies = [[b.name,float(b.scf_done[-1][1]),b.normal_termin,b.last_xyz_obj()] for i in file_weeder([".log"]) for b in [LogFile(read_item(i))]]
	unique = energies
	if not unique: print("No log files to be analyzed"); return
	black_list, folder_mov = [], []
	print("Starting analysis...")
	for file in sorted(unique,key=lambda x: (x[2],-x[1]),reverse=True):
		if file[0] in black_list: continue
		black_list.append(file[0])
		sim_en = [i for i in unique if i[0] not in black_list and  i[1] + 1 > file[1] > i[1] - 1]
		if sim_en:
			duplicates = []
			for obj in sim_en:
				if obj[3].superimpose(file[3], ret="max_d",conv=6):
					print("{} is a duplicate of {}".format(obj[3].name(), file[3].name()))
					duplicates.append(obj[3].name())
					black_list.append(obj[3].name())
			if duplicates: folder_mov.append([file[3].name(),duplicates])
	for folder in folder_mov:
		subfolder = os.path.join(os.getcwd(),"duplicates_of_{}".format(folder[0].replace(".log","")))
		try: os.mkdir(subfolder)
		except FileExistsError:	pass
		print("Moving duplicates of {} to the following directory:\n{}".format(folder[0], subfolder))
		for file in folder[1]:
			for alt_file in file_weeder([file.replace(".log",".")]):
				try:
					shutil.move(os.path.join(os.getcwd(),alt_file),os.path.join(subfolder,alt_file))
					print("Moved: {}".format(alt_file))
				except PermissionError:
					print("Error while moving log files:")
					print("Maybe {} already exists in the following directory:\n{}".format(alt_file,subfolder))
				except FileNotFoundError:
					print("File {} not found in the following directory:\n{}".format(alt_file,subfolder))
	print("Done!")



import os
from raachem.util.gen_purp import file_weeder, preferences

def deploy():
	scripts_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)),"scripts")
	while True:
		options = file_weeder([".py"], scripts_dir)
		options = [a for a in options if a != "__init__.py"]
		options.insert(0, "To cancel")
		print("Which script do you want to paste in the current folder?")
		for idx, i in enumerate(options):
			print("{:>3} - {}".format(idx, i))
		option = input()
		if option in [str(a) for a in range(len(options))]:
			break
	if option == "0":
		return
	else:
		with open(os.path.join(scripts_dir,options[int(option)])) as file:
			script = file.readlines()
		script = [a.replace("sub_s_name",preferences.sub_s_name) for a in script]
		script = [a.replace(".gjf", preferences.gauss_ext) for a in script]
		with open((os.path.join(os.getcwd(), options[int(option)])), "w", newline="\n") as file_b:
			for line in script:
				file_b.write(str(line))



from raachem.file_class.gjf import *
from raachem.file_class.inp import *
from raachem.file_class.xyz import *
from raachem.file_class.log import *
from raachem.file_creator.e_analysis import *
from raachem.file_creator.input import *
from raachem.file_creator.xyz import *
from raachem.file_creator.deploy_scripts import *
from raachem.util.gen_purp import *




############################RAACHEM IMPORTS###############################
from raachem.file_creator.e_analysis import e_analysis, csv_e_analysis, deduplicate
from raachem.file_creator.input import CreateInputs, xyz_insert, validate_input
from raachem.file_creator.xyz import input_to_xyz, log_to_xyz, log_to_xyz_scan, log_freq_xyz, superimpose_alg, geodes_int, xyz_ent
from raachem.file_creator.deploy_scripts import deploy
from raachem.util.gen_purp import file_weeder, mv_up_folder, preferences
###############################INTERFACE##################################

class UserInterface:
	def __init__(self):
		self.menu_a = [self.func_00,self.func_01,self.func_02,
					   self.func_03,self.func_04,self.func_05,
					   self.func_06,self.func_07,self.func_21,
					   self.func_09]
		self.menu_b = [self.func_10,self.func_12,self.func_13,
					   self.func_14,self.func_15,self.func_16,
					   self.func_17,self.func_18,self.func_20,
					   self.func_19,]
		try:
			import raapbs
			self.pbs = True
		except ImportError:
			self.pbs = False
	def describe(self):
		"""Call help for all functions"""
		for a in dir(self):
			if a.startswith("func"):
				help(f"raachem.util.interface.UserInterface.{a}")
	def render(self,menu=None):
		"""Render menu"""
		if menu is None: menu = self.menu_a
		while True:
			if preferences.comp_software == "gaussian":	print("Choose an option:")
			else: print("Choose an option (option* = functions for gaussian software files):")
			for i,a in enumerate(menu):	a(mode="print",idx=i)
			while True:
				option = input()
				if option in (str(a) for a in range(len(menu))):break
				else: print("Invalid input. Could not process request!")
			menu[int(option)](mode="run")
	@staticmethod
	def func_00(mode="print",idx=None):
		"""Exits the program"""
		if mode == "print": print(f"{idx} - Exit")
		elif mode == "run": exit(print("ok"))
	@staticmethod
	def func_01(mode="print",idx=None):
		"""Create .xyz files from log files"""
		if preferences.comp_software == "orca":
			if mode == "print":	print(f"{idx} - Create .xyz files from .log files*")
			elif mode == "run":	log_to_xyz()
		elif preferences.comp_software == "gaussian":
			if mode == "print":	print(f"{idx} - Create .xyz files from .log files")
			elif mode == "run": log_to_xyz()
	@staticmethod
	def func_02(mode="print",idx=None):
		"""Create input files from .xyz files"""
		if preferences.comp_software == "orca":
			if mode == "print":	print(f"{idx} - Create .inp files from .xyz files")
			elif mode == "run":	CreateInputs()
		elif preferences.comp_software == "gaussian":
			if mode == "print":	print(f"{idx} - Create {preferences.gauss_ext} files from .xyz files")
			elif mode == "run":	CreateInputs()
	def func_03(self,mode="print",idx=None):
		"""Validade inputs and possibly create pbs files"""
		if preferences.comp_software == "orca":
			if mode == "print":
				if self.pbs: print(f'{idx} - Create .pbs files and a submission script from .inp files')
				else: print(f'{idx} - Validate .inp files')
			elif mode == "run":
				validate_input(file_weeder([".inp"]))
		elif preferences.comp_software == "gaussian":
			if mode == "print":
				if self.pbs: print(f"{idx} - Create .pbs files and a submission script from {preferences.gauss_ext} files")
				else: print(f"{idx} - Validate {preferences.gauss_ext} files")
			elif mode == "run":	validate_input(file_weeder([preferences.gauss_ext]))
	@staticmethod
	def func_04(mode="print",idx=None):
		"""Analyze scan, opt and irc files"""
		if preferences.comp_software == "orca":
			if mode == "print":	print(f"{idx} - IRC, OPT or SCAN analysis*")
			elif mode == "run":	log_to_xyz_scan()
		elif preferences.comp_software == "gaussian":
			if mode == "print": print(f"{idx} - IRC, OPT or SCAN analysis")
			elif mode == "run":	log_to_xyz_scan()
	@staticmethod
	def func_05(mode="print",idx=None):
		"""Insert xyz geometries into input"""
		if preferences.comp_software == "orca":
			if mode == "print":	print(f"{idx} - Insert .xyz geometries into the corresponding .inp files")
			elif mode == "run":	xyz_insert(file_weeder([".xyz"]))
		elif preferences.comp_software == "gaussian":
			if mode == "print":	print(f"{idx} - Insert .xyz geometries into the corresponding {preferences.gauss_ext} files")
			elif mode == "run":	xyz_insert(file_weeder([".xyz"]))
	@staticmethod
	def func_06(mode="print",idx=None):
		"""Analyze relative free energies"""
		if preferences.comp_software == "orca":
			if mode == "print":	print(f"{idx} - Extract free energies from .log files*")
			elif mode == "run": e_analysis(file_weeder([".log"]))
		elif preferences.comp_software == "gaussian":
			if mode == "print":	print(f"{idx} - Extract free energies from .log files")
			elif mode == "run": e_analysis(file_weeder([".log"]))
	@staticmethod
	def func_07(mode="print",idx=None):
		"""Create inputs from .log files"""
		if preferences.comp_software == "orca":
			if mode == "print":	print(f"{idx} - Create .inp files from .log files")
			elif mode == "run":	CreateInputs(use_logs=True)
		elif preferences.comp_software == "gaussian":
			if mode == "print":	print(f"{idx} - Create {preferences.gauss_ext} files from .log files")
			elif mode == "run":	CreateInputs(use_logs=True)
	def func_08(self,mode="print",idx=None):
		"""Placeholder"""
		_ , _ , _ = self, mode, idx
	def func_09(self,mode="print",idx=None):
		"""More options"""
		if mode == "print":	print(f"{idx} - More options")
		elif mode == "run":	self.render(self.menu_b)
	def func_10(self,mode="print",idx=None):
		"""Go back to previous menu"""
		if mode == "print":	print(f"{idx} - Go back to previous menu")
		elif mode == "run":	self.render(self.menu_a)
	@staticmethod
	def func_11(mode="print",idx=None):
		"""Create energy profile as .svg"""
		if mode == "print":	pass
		elif mode == "run":	pass
	@staticmethod
	def func_12(mode="print",idx=None):
		"""Return enantiomer of xyz file"""
		if mode == "print":	print(f"{idx} - Create enantiomer of xyz files")
		elif mode == "run":	xyz_ent()
	@staticmethod
	def func_13(mode="print",idx=None):
		"""Move files up a folder"""
		if mode == "print":	print(f"{idx} - Move Files up a folder")
		elif mode == "run":	mv_up_folder()
	@staticmethod
	def func_14(mode="print",idx=None):
		"""Create to xyzs from a given vibration"""
		if preferences.comp_software == "orca":
			if mode == "print":	print(f"{idx} - Split frequency in two directions and create corresponding .xyzs*")
			elif mode == "run":	log_freq_xyz()
		elif preferences.comp_software == "gaussian":
			if mode == "print":	print(f"{idx} - Split frequency in two directions and create corresponding .xyzs")
			elif mode == "run":	log_freq_xyz()
	@staticmethod
	def func_15(mode="print",idx=None):
		"""Superimpose xyz structures"""
		if mode == "print":	print(f"{idx} - Superimpose two xyz files")
		elif mode == "run":	superimpose_alg()
	@staticmethod
	def func_16(mode="print",idx=None):
		"""Generate solvated .xyzs"""
		if mode == "print":	print(f"{idx} - Generate solvated .xyz's")
		elif mode == "run":	geodes_int()
	@staticmethod
	def func_17(mode="print",idx=None):
		"""Deduplicate .log files"""
		if preferences.comp_software == "orca":
			if mode == "print":	print(f"{idx} - Deduplicate .log files*")
			elif mode == "run":	deduplicate()
		elif preferences.comp_software == "gaussian":
			if mode == "print":	print(f"{idx} - Deduplicate .log files")
			elif mode == "run":	deduplicate()
	@staticmethod
	def func_18(mode="print",idx=None):
		"""Create .xyz files from input files"""
		if preferences.comp_software == "orca":
			if mode == "print":	print(f"{idx} - Create .xyz files from .inp files")
			elif mode == "run":	input_to_xyz()
		elif preferences.comp_software == "gaussian":
			if mode == "print":	print(f"{idx} - Create .xyz files from {preferences.gauss_ext} files")
			elif mode == "run":	input_to_xyz()
	@staticmethod
	def func_19(mode="print",idx=None):
		"""Analyze log files in current and sub directories"""
		if preferences.comp_software == "orca":
			if mode == "print":	print(f"{idx} - Analyze log files in current and sub directories*")
			elif mode == "run":	csv_e_analysis()
		elif preferences.comp_software == "gaussian":
			if mode == "print":	print(f"{idx} - Analyze log files in current and sub directories")
			elif mode == "run":	csv_e_analysis()
	@staticmethod
	def func_20(mode="print",idx=None):
		"""Configure"""
		if mode == "print":	print(f"{idx} - Configure")
		elif mode == "run":	preferences.set_variables()
	@staticmethod
	def func_21(mode="print",idx=None):
		"""Deploy scripts"""
		if mode == "print":	print(f"{idx} - Deploy script")
		elif mode == "run":	deploy()




from raachem.util.constants import elements
from raachem.file_class.xyz import XyzFile
import functools
class InpFile:
	possible_blocks = ("autoci", "basis", "casscf", "cipsi", "cis", "cim", "coords", "cpcm", "elprop", "eprnmr",
					   "freq", "geom", "loc", "md", "mdci", "method", "mp2", "mrci", "mrcc", "numgrad", "nbo",
					   "output", "pal", "paras", "rel", "plots", "rocis", "rr", "scf", "vpt2")

	def __init__(self,file_content):
		self.list = file_content
		self.list_l = [a.split() for a in file_content]
		self.str_l = [a.replace(" ","") for a in self.list]
		self.return_print = "\n".join(self.list[1:])
		#########################
		#########################
		self.comt_ls = [i for i,a in enumerate(self.str_l) if a.startswith("#")]
		self.keys_ls = [i for i,a in enumerate(self.str_l) if a.startswith("!")]
		self.ljob_ls = [i for i,a in enumerate(self.str_l) if a.startswith("%base")]
		self.njob_ls = [i for i,a in enumerate(self.str_l) if a.startswith("$new_job")]
		self.file_ls = [i for i,a in enumerate(self.str_l) if a.startswith("*xyzfile")]
		self.astk_ls = [i for i,a in enumerate(self.str_l) if a == "*"]
		self.cord_ls = [i for i,a in enumerate(self.str_l) if a.startswith("*xyz") and not i in self.file_ls]
		self.perc_ls = [i for i,a in enumerate(self.str_l) if a.startswith("%") and not i in self.ljob_ls]
		self.ends_ls = [i for i,a in enumerate(self.str_l) if a == "end"]
	@functools.lru_cache(maxsize=1)
	def name(self):
		if len(self.list[0]) == 0: raise Exception(".inp object has no name")
		return self.list[0]
	@functools.lru_cache(maxsize=1)
	def charge(self):
		return self._charge_multiplicity()[0]
	@functools.lru_cache(maxsize=1)
	def multiplicity(self):
		return self._charge_multiplicity()[1]
	@functools.lru_cache(maxsize=1)
	def n_electrons(self):
		return sum(elements.index(e) for e in self.all_elements()) - self.charge()
	@functools.lru_cache(maxsize=1)
	def n_atoms(self):
		return len(self.all_elements())
	@functools.lru_cache(maxsize=1)
	def all_elements(self):
		return [a[0] for a in self.list_l[self.c_m_idx()+1:self.end_cord_idx()]] if self.cord_ls else None
	@functools.lru_cache(maxsize=1)
	def elements(self):
		return list(dict.fromkeys(self.all_elements()))
	@functools.lru_cache(maxsize=1)
	def c_m_validate(self):
		return not self.n_electrons()%2 == self.multiplicity()%2
	@functools.lru_cache(maxsize=1)
	def c_m_validate_txt(self):
		return "Yes" if self.c_m_validate() else "--NO!--"
	@functools.lru_cache(maxsize=1)
	def n_proc(self):
		n_proc = None
		for a in self.route_text().split():
			if not a.lower().startswith("pal"):continue
			if all(b.isdigit() for b in a.lower()[3:]): n_proc = int(a.lower()[3:])
		if n_proc == 1: print("ALERT: For single core jobs the number of parallel processors 'PAL' should be omited!")
		while True:
			if "pal" not in self.block_keys(): break
			if "nprocs" not in self.block_keys()["pal"]: break
			print("WARNING: Proccessor count seems to have been provided twice: {}".format(self.name))
			if self.block_keys()["pal"].index("nprocs") + 1  == len(self.block_keys()["pal"]): break
			value = self.block_keys()[self.block_keys()["pal"].index("nprocs") + 1]
			if value.isdigit(): n_proc = int(value)
			break
		if n_proc is None: n_proc = 1
		return n_proc
	@functools.lru_cache(maxsize=1)
	def cord_block(self):
		return [*self.list_l[self.c_m_idx()+1:self.end_cord_idx()]]
	@functools.lru_cache(maxsize=1)
	def route_text(self):
		return " ".join([self.list[a].replace("!", "", 1) for a in self.keys_ls])
	@functools.lru_cache(maxsize=1)
	def c_m_idx(self):
		return min(self.cord_ls)
	@functools.lru_cache(maxsize=1)
	def end_cord_idx(self):
		return min(a for a in self.astk_ls if a > self.c_m_idx())
	#########################
	#########################
	@functools.lru_cache(maxsize=1)
	def _charge_multiplicity(self):
		try:
			charge = int(self.list[self.c_m_idx()].split()[-2])
			mult = int(self.list[self.c_m_idx()].split()[-1])
		except IndexError:
			charge = None
			mult = None
			print("Did you provide charge and multiplicity data?")
			print("Line {}:\n'{}'".format(self.c_m_idx()+1,self.list[self.c_m_idx()]))
		except ValueError:
			charge = None
			mult = None
			print("Did you properly provide charge and multiplicity data?")
			print("Line {}:\n'{}'".format(self.c_m_idx()+1,self.list[self.c_m_idx()]))
		finally:
			return [charge,mult]
	@functools.lru_cache(maxsize=1)
	def block_keys(self,possible_blocks=possible_blocks):
		d = {}
		for a in self.perc_ls:
			first_key = self.list[a].replace("%","",1).split()
			while True:
				if len(first_key) == 0: a += 1; first_key = self.list[a].split()
				else: break
			if first_key[0].lower() in possible_blocks:
				end_line = self._hunt_end(a)
				if end_line != None: d[first_key[0].lower()] = " ".join(self.list[a:end_line+1]).replace("%","",1).split()
		return d

	#########################
	#########################
	def replace_cord(self,xyz_obj):
		new = []
		for line in self.list[0:self.c_m_idx()+1]: new.append(line)
		for line in xyz_obj.form_cord_block(): new.append(line)
		for line in self.list[self.end_cord_idx():]: new.append(line)
		return InpFile(new)
	def _hunt_end(self,start):
		for idx,line in enumerate(self.list[start:],start=start):
			if "end" in line.lower().split(): return idx
		return None
	def xyz_obj(self):
		return XyzFile([self.name(),self.n_atoms()," ",*[" ".join(a) for a in self.cord_block()]])






#!/usr/bin/env python
import os
current_folder=str(os.path.dirname(os.path.abspath(__file__)))

def evaluate_list(folder,opt_a,scripts=[]):
	exetensions = [".xyz",".chk",".GAUSSIAN"]
	a_options = ["2","3","4"]
	for file in os.listdir(folder):
		if os.path.isdir(os.path.join(folder,file)):
			if os.path.join(folder,file).endswith("miniconda3"):continue
			if os.path.join(folder,file).endswith("bin"):continue
			evaluate_list(os.path.join(folder,file),opt_a,scripts)
			continue
		try:
			ext = os.path.splitext(file)[1]
			if any(ext == a for a,b in zip(exetensions,a_options) if b in opt_a):
				print(os.path.join(folder,file))
				scripts.append(os.path.join(folder,file))
			if "1" not in opt_a: continue
			if len(ext) != 7: continue
			if not ext[1].isalpha():continue
			if not ext[1].islower():continue
			if not all(a.isdigit() for a in ext[2:]):continue
			else:print(os.path.join(folder,file)) 
			scripts.append(os.path.join(folder,file))
		except IndexError:
			print(file," has no extension!")
			continue
	return	scripts
print("What file types do you want to remove based on their extension?")
print("0 - To return without delleting files")
print("1 - Files with the '.XYYYYY' extension where X are letters and Y are numbers")
print("2 - Files with the '.xyz' extension")
print("3 - Files with the '.chk' extension")
print("4 - Files with the '.GAUSSIAN' extension")
while True:
	opt_a = input().split()
	if all(a in ["0","1","2","3","4"] for a in opt_a):break
	else: print("Invalid Input, please try again")
if "0" in opt_a: exit()
print("Files to be removed:")
filename = evaluate_list(current_folder, opt_a)
print("Do you want to remove the above files?(y/n)")
option = input()
if option == "y":
	for file in filename:
		os.system("rm {0}".format(file))
	print("Done!")
else:
	print("No files will be deleted")




import os, math, shutil, re
from difflib import SequenceMatcher
from raachem.file_class.xyz import XyzFile
from raachem.file_class.gjf import GjfFile
from raachem.file_class.log import LogFile
from raachem.file_class.inp import InpFile
from raachem.util.constants import element_radii, keywords, elements
from raachem.util.gen_purp import file_weeder, read_item, preferences, sel_files, is_str_float

class CreateInputs:
	def __init__(self,use_logs=False):
		self.use_logs = use_logs
		self.original_ext = ".log" if use_logs else ".xyz"
		self.template_ext = ".GAUSSIAN" if preferences.comp_software == "gaussian" else ".ORCA"
		self.extension = preferences.gauss_ext if preferences.comp_software == "gaussian" else ".inp"
		self.weeded_list = file_weeder([self.original_ext])
		self.p_files_dir = os.path.join(os.path.dirname(__file__),"builtin_templates")
		self.user_files_dir = os.path.join(os.path.dirname(__file__), "par_templates")
		self.submit_parameters()
	def submit_parameters(self):
		l_files = file_weeder([self.template_ext])
		if len(l_files) == 1:
			print("Reading parameters from {} in current folder".format(l_files[0]))
			parameters = read_item(l_files[0])
			self.find_idx(parameters)
		elif len(l_files) > 1:
			l_files.insert(0, None)
			option = True
			print("Chose a parameter file:")
			for idx, i in enumerate(l_files):
				if i is None: print("  0 - Cancel")
				else: print("{:>3} - {}".format(idx, i))
			while option == True: option = {str(idx): i for idx, i in enumerate(l_files)}.get(input(), True)
			if option is None: return
			else: parameters = read_item(option); self.find_idx(parameters)
		else:
			print("Could not find {} file on current file".format(self.template_ext))
			option = True
			while option == True:
				print("Chose an option:")
				print("0 - To cancel")
				print("1 - Use builtins")
				print("2 - User defined")
				option = {str(i): str(i) for i in range(3)}.get(input(), True)
			if option == "0": return
			else:
				folder = self.p_files_dir if option == "1" else self.user_files_dir
				p_files = file_weeder([self.template_ext], cf=folder)
				p_files.insert(0, None)
				print(f"Reading {self.template_ext} files from:\n{folder}\nChoose a parameter file:")
				option_2 = True
				for idx, i in enumerate(p_files):
					if i == None: print("  0 - Cancel")
					else: print("{:>3} - {}".format(idx, i))
				while option_2 == True: option_2 = {str(idx): i for idx, i in enumerate(p_files)}.get(input(), True)
				if option_2 == None: return
				else:
					shutil.copy(os.path.join(folder, option_2), os.path.join(os.getcwd(), "PARAMETERS{}".format(self.template_ext)))
					self.try_again()
	def try_again(self):
		print("A PARAMETERS{} file was added to the current folder!".format(self.template_ext))
		print("You can edit it now")
		print("0 - To cancel")
		print("1 - File is ready!")
		option = True
		while option == True: option = {str(i): str(i) for i in range(2)}.get(input(), True)
		if option == "0": return
		elif option == "1":	self.__init__(use_logs=self.use_logs)
	def find_idx(self,parameters):
		lookup_str = "=====WILL BE GEOMETRY BLOCK====="
		for idx, line in enumerate(parameters[1:]):
			if line.strip() == lookup_str and self.template_ext == ".GAUSSIAN": self.save_gjf(parameters[1:], idx); return
			if line.strip() == lookup_str and self.template_ext == ".ORCA": self.save_inp(parameters[1:], idx); return
		print("The following line is required on the {} file:".format(self.template_ext))
		print(lookup_str)
		return
	def save_inp(self,parameters,index):
		if not preferences.folder_op: self.weeded_list = sel_files(self.weeded_list)
		if not self.weeded_list: return
		for i in self.weeded_list:
			if os.path.isfile(os.path.join(os.getcwd(), (i.replace(self.original_ext, ".inp")))):
				print(i.replace(self.original_ext,".inp") + " already exist on current directory!")
				continue
			xyz = LogFile(read_item(i)).last_xyz_obj() if self.use_logs else XyzFile(read_item(i))
			inp_out = []
			for idx_a, line in enumerate(parameters[0:index]):
				if self.use_logs and idx_a + 1 == len(parameters[0:index]):
					line = "* " + " ".join(LogFile(read_item(i)).charge_mult)
				inp_out.append(line.replace("FILENAME", i.replace(self.original_ext, "")) + "\n")
			for line in xyz.form_cord_block():
				inp_out.append(line + "\n")
			for line in parameters[index + 1:]:
				inp_out.append(line + "\n")
			with open(os.path.join(os.getcwd(), (i.replace(self.original_ext,".inp"))), "w",newline="\n") as file:
				for line in inp_out: file.write(line)
			print(i.replace(self.original_ext, ".inp"), " created!")
		return
	def save_gjf(self,parameters,index):
		heavy_e, gjf_overwrite, folder_op = [preferences.heavy_atom,preferences.gjf_overwrite,preferences.folder_op]
		ecps, basis = None, None
		if not folder_op: self.weeded_list = sel_files(self.weeded_list)
		if not self.weeded_list: return
		if any([b in parameters for b in ["INSERT_GBS_BASIS","INSERT_GBS_ECP"]]):
			print("This parameter file requires an aditional gbs file.")
			print("Where should it be taken from?")
			print("0 - Current folder")
			print("1 - Builtins")
			print("2 - User defined")
			while True:
				option = input()
				if option in ["0","1","2"]: break
			if option == "0": files_dir = os.getcwd()
			elif option == "1": files_dir = self.p_files_dir
			elif option == "2": files_dir = self.user_files_dir
			gbs_files = file_weeder([".gbs"], cf=files_dir)
			print("Chosse one basis/ecp set\n0 - Cancel")
			for i, file in enumerate(gbs_files): print("{} - {}".format(i+1, file))
			while True:
				option = input()
				if option in [str(a) for a in range(len(gbs_files)+1)]: break
				print("Invalid option")
			if option == "0": return
			with open(os.path.join(files_dir,gbs_files[int(option)-1])) as file:
				gbs_file = file.read().splitlines()
			basis, ecps = self.read_gbs(gbs_file)
			assert type(basis) == dict, "Basis was not read"
			assert type(ecps) == dict, "Ecps were not read"
		for i in self.weeded_list:
			try:
				gjf_out, rm_lines, = [], []
				if os.path.isfile(os.path.join(os.getcwd(),(i.replace(self.original_ext,preferences.gauss_ext)))) and not gjf_overwrite:
					print(i.replace(self.original_ext,preferences.gauss_ext) + " already exist on current directory!")
					continue
				xyz = LogFile(read_item(i)).last_xyz_obj() if self.use_logs else XyzFile(read_item(i))
				for idx_a,line in enumerate(parameters[0:index]):
					if self.use_logs and idx_a+1 == len(parameters[0:index]): line = " ".join(LogFile(read_item(i)).charge_mult)
					gjf_out.append(line.replace("FILENAME",i.replace(self.original_ext,""))+"\n")
				for line in xyz.form_cord_block():
					gjf_out.append(line+"\n")
				possible = [str(b + 1) for b in range(xyz.n_atoms())]
				for line in parameters[index+1:]:
					# SCAN like: "B S APROX" or "B S DIST"
					is_mod_red = any("modredundant" in i.lower() for i in parameters[0:index])
					s_line = line.split()
					while len(s_line) == 3:
						if s_line[0] != "B" or s_line[1] != "S": break
						if s_line[2] not in ["APROX","DIST"]: break
						if not is_mod_red:
							print("Missing 'modredundant' keyword?")
							break
						if xyz.n_atoms() <2:
							raise Exception(f"At least two atoms are neded in struture {xyz.name()} to perform a scan")
						atom=["a","b"]
						if s_line[-1] == "APROX":
							print(f"Detected bond scan (APROX) for xyz file {i}.")
							while not all(atom[a] in possible if atom[0] != atom[1] else False for a in [0,1]):
								atom = [input("Enter atom A: "),input("Enter atom B: ")]
							line = f"B {atom[0]} {atom[1]} S APROX"
						elif s_line[-1] == "DIST":
							print(f"Detected bond scan (DIST) for xyz file {i}.")
							while not all(atom[a] in possible if atom[0] != atom[1] else False for a in [0,1]):
								atom = [input("Enter atom A: "),input("Enter atom B: ")]
							line = f"B {atom[0]} {atom[1]} S DIST"
						s_line = line.split()
						break
					# SCAN like: "B 1 2 S APROX" or "B 1 2 S DIST"
					while len(s_line) == 5:

						if s_line[0] != "B" or s_line[3] != "S": break
						if not s_line[1].isdigit(): break
						if not s_line[2].isdigit(): break
						if s_line[4] not in ["APROX", "DIST"]: break
						if not is_mod_red:
							print("Missing 'modredundant' keyword?")
							break
						if xyz.n_atoms() < 2:
							raise Exception(f"At least two atoms are neded in struture {xyz.name()} to perform a scan")
						atoms_idx = [int(s_line[1])-1, int(s_line[2])-1]
						if any(True for a in atoms_idx if not a in range(xyz.n_atoms())):
							raise Exception(f"Scan atom numbers are larger than the number of atoms for: {xyz.name()}")
						atoms_cord = [i for idx,i in enumerate(xyz.cord_block()) if idx in atoms_idx]
						dist = math.sqrt(sum((float(i)-float(atoms_cord[1][idx]))**2 for idx,i in enumerate(atoms_cord[0]) if idx > 0))
						ideal_dist=sum(b[1] for idx,b in enumerate(element_radii) if b[0] in [i[0] for i in atoms_cord])/100
						if s_line[-1] == "APROX":
							line = line.replace("APROX",f"{1+int((dist-ideal_dist)/0.075)} -0.075")
						elif s_line[-1] == "DIST":
							line = line.replace("DIST",f"{1+int(ideal_dist/0.075)} 0.075")
						s_line = line.split()
						break
					# SCAN like: "D S"
					while len(s_line) == 2:
						if not is_mod_red: break
						if s_line[0] != "D" or s_line[1] != "S": break
						print(f"Detected dihedral scan for xyz file {i}.")
						if xyz.n_atoms() < 4:
							raise Exception(f"At least two four atoms are neded in struture {xyz.name()} to perform a dihedral scan")
						while True:
							atom = [input(f"Enter atom {a}: ") for a in ["A","B","C","D"]]
							atom = [a.strip() for a in atom]
							if not all([len(a.split()) == 1 and a.isdigit() for a in atom]):
								print("No non-numeric characters or spaces are allowed for atom numbers")
								continue
							if not len(set(atom)) == 4:
								print("No atom should be repeated!")
								continue
							if not all([a in possible for a in atom]):
								print("Atoms with incorrect number were found!")
								continue
							break
						while True:
							n_steps = input("Please give the number of steps to be taken: ").strip()
							if n_steps.isdigit(): break
							else: print("The number of steps must be an integer")
						while True:
							print("Please give the step size to be taken in degrees")
							size = input("The number must include a dot character as a decimal place (eg. '10.0'): ").strip()
							if is_str_float(size) and size.count(".") == 1: break
							else: print("Please make sure the number complies with formating rules!")
						line = f"D {' '.join(atom)} S {n_steps} {size}"
						s_line = line.split()
						break
					# READOPTIMIZE + NOTATOMS
					if line.replace(" ", "").lower() == "notatoms=":
						print("Please enter the atoms you want to freeze on structure {} separated by comma".format(xyz.name()))
						line = line + input(line)
					# READOPTIMIZE + ATOMS
					elif line.replace(" ", "").lower() == "noatomsatoms=":
						print("Please enter the atoms you want to optimize on structure {} separated by comma".format(xyz.name()))
						line = line + input(line)
					# BASIS like: "LIGHT_ELEMENT_BASIS 0" or "HEAVY_ELEMENT_BASIS 0" and ECP like "HEAVY_ELEMENT_ECP 0"
					elif len(s_line) == 2:
						if any(True for a in ["/gen","gen ","genecp"] if a in " ".join(parameters[0:index-3]).lower()):
							if s_line == ["LIGHT_ELEMENT_BASIS","0"]:
								elm = [a for a in xyz.elements() if elements.index(a) < heavy_e+1]
								if elm: line = line.replace("LIGHT_ELEMENT_BASIS"," ".join(elm))
								else:
									for a in range(3): rm_lines.append(len(gjf_out)+a)
							elif s_line == ["HEAVY_ELEMENT_BASIS","0"]:
								elm = [a for a in xyz.elements() if elements.index(a) > heavy_e]
								if elm: line = line.replace("HEAVY_ELEMENT_BASIS"," ".join(elm))
								else:
									for a in range(3): rm_lines.append(len(gjf_out)+a)
						if "pseudo=read" in " ".join(parameters[0:index - 3]).lower().replace(" ",""):
							if s_line == ["HEAVY_ELEMENT_ECP","0"]:
								elm = [a for a in xyz.elements() if elements.index(a) > heavy_e]
								if elm: line = line.replace("HEAVY_ELEMENT_ECP"," ".join(elm))
								else:
									pattern = re.compile(r'pseudo.{0,3}=.{0,3}read',re.IGNORECASE)
									eval_lines = gjf_out[0:index-3]
									for idx_a,a in enumerate(eval_lines):
										gjf_out[idx_a] = pattern.sub("",a)
									rm_lines.append(len(gjf_out))
									rm_lines.append(len(gjf_out)+1)
					# BASIS like: "INSERT_GBS_BASIS"
					elif "INSERT_GBS_BASIS" in line:
						for element in sorted(xyz.elements(),key=lambda x: elements.index(x)):
							for line_a in basis[element]:
								gjf_out.append(line_a+"\n")
						continue
					# ECP like: "INSERT_GBS_ECP"
					elif "INSERT_GBS_ECP" in line:
						if "pseudo=read" in " ".join(parameters[0:index - 3]).lower().replace(" ", ""):
							need_ecp = [a for a in xyz.elements() if elements.index(a) > preferences.heavy_atom]
							for element in sorted(need_ecp,key=lambda x: elements.index(x)):
								for line_a in ecps[element]:
									gjf_out.append(line_a+"\n")
							if not need_ecp:
								pattern = re.compile(r'pseudo.{0,3}=.{0,3}read', re.IGNORECASE)
								eval_lines = gjf_out[0:index - 3]
								for idx_a, a in enumerate(eval_lines):
									gjf_out[idx_a] = pattern.sub("", a)
							continue
					gjf_out.append(line.replace("FILENAME",i.replace(self.original_ext,""))+"\n")
				gjf_out.append("\n")
				with open(os.path.join(os.getcwd(),(i.replace(self.original_ext,preferences.gauss_ext))),"w") as gjf_file:
					for line in [i for idx,i in enumerate(gjf_out) if idx not in rm_lines]: gjf_file.write(line)
				print(i.replace(self.original_ext,preferences.gauss_ext)," created!")
			except TypeError as e:
				print("Error on file {}".format(i))
				print(e)
		return
	@staticmethod
	def read_gbs(gbs_file):
		gbs = [a for a in gbs_file if not a.startswith("!")]
		gbs_starts = []
		for i,a in enumerate(gbs):
			if len(a.split()) != 2: continue
			if not a.split()[0].capitalize() in elements: continue
			if a.split()[1] != "0":continue
			else: gbs_starts.append(i)
		gbs_basis_ends = [i+1 for i,a in enumerate(gbs) if a.startswith("****")]
		basis_dict = {gbs[a].split()[0].capitalize():gbs[a:b] for a,b in zip(gbs_starts,gbs_basis_ends) if a < b}
		ecp_starts = [a for a in gbs_starts if a > max(gbs_basis_ends)]
		ecp_ends = []
		for start in ecp_starts:
			for i,line in enumerate(gbs[start:]):
				if i < 1:continue
				elif len(line.split()) == 0: ecp_ends.append(i+start); break
				elif i + 1 == len(gbs[start:]):  ecp_ends.append(i+start); break
				elif len(line.split()) != 2: continue
				elif line.split()[0].capitalize() in elements and line.split()[1] == "0": ecp_ends.append(i+start); break
		ecp_dict = {gbs[a].split()[0].capitalize():gbs[a:b] for a,b in zip(ecp_starts,ecp_ends) if a < b}
		return basis_dict,ecp_dict
def xyz_insert(weeded_list):
	"""Inserts geometries into both orca and gaussian input files"""
	extension = ".inp" if preferences.comp_software == "orca" else preferences.gauss_ext
	sub_folder = os.path.join(os.getcwd(),"inserted_input_files")
	if os.path.exists(sub_folder):
		print("'inserted_input_files' directory already exists in current directory!")
		print("Please remove it and try again!")
		return
	os.mkdir(sub_folder)
	for i in weeded_list:
		try:
			xyz = XyzFile(read_item(i))
			if preferences.comp_software == "orca": comp_input = InpFile(read_item(i.replace(".xyz",extension)))
			else: comp_input = GjfFile(read_item(i.replace(".xyz",extension)))
			comp_input = comp_input.replace_cord(xyz)
			with open(os.path.join(sub_folder,i.replace(".xyz",extension)),"w") as file:
				file.write(comp_input.return_print)
		except FileNotFoundError: print("file " + i.replace(".xyz",extension) + " could not be found!")
	print("\nJob done!\nPlease lookup the inserted_input_files directory\n")
	return

def validate_input(weeded_list):
	print("---------------------------------------------------------------------------")
	print("{:^30}{:^15}{:^10}{:^10}{:^10}".format("File","e- number","charge","multip","Validated"))
	print("---------------------------------------------------------------------------")
	novel_keys = []
	for item in weeded_list:
		if preferences.comp_software == "orca":	comp_input = InpFile(read_item(item))
		else: comp_input = GjfFile(read_item(item))
		a = (comp_input.name(), comp_input.n_electrons(), comp_input.charge(), comp_input.multiplicity(), comp_input.c_m_validate_txt())
		print(" {:<30}{:>10}{:>10}{:>10}{:^15}".format(*a))
		if preferences.comp_software == "gaussian":
			split_list = [i for i in comp_input.list[1:comp_input.title_idx()] if not i.lower().startswith("%chk")]
			for x in [None,"/","(",")",",","=","%",":"]:
				split_list = [a for b in [i.split(x) for i in split_list] for a in b if len(a) > 3]
			no_match = [i for i in split_list if i.lower() not in [j.lower() for j in keywords] and not i[0].isdigit()]
			typo_error = []
			for b in no_match:
				p_matches = [[b,i,SequenceMatcher(None,b,i).ratio()] for i in keywords if SequenceMatcher(None,b,i).ratio() > 0.6]
				if p_matches:
					p_matches.sort(key=lambda x: x[2])
					typo_error.append([p_matches[0][0],p_matches[0][1]])
			novel_keys.append([i for i in no_match if i not in [a[0] for a in typo_error]])
			if any([comp_input.route_errors(),comp_input.basis_errors(),typo_error,comp_input.ecp_errors(preferences.heavy_atom),len(comp_input.name().split()) != 1]):
				print("{:>7}+----------------------------ALERT---------------------------+".format(" "))
				for error in comp_input.basis_errors(): print("{:>8}{:>60}".format("|",error+" |"))
				for error in comp_input.ecp_errors(preferences.heavy_atom): print("{:>8}{:>60}".format("|",error+" |"))
				for error in comp_input.route_errors(): print("{:>8}{:>60}".format("|", error + " |"))
				for i in typo_error: print("{:>8}{:>60}".format("|", "Is '{}' a typo of '{}'?".format(i[0],i[1]) + " |"))
				if len(comp_input.name().split()) != 1: print("{:>8}{:>60}".format("|", "Filename must not contain spaces!" + " |"))
				print("{:>7}+-----------------------------------------------------------+".format(" "))
	print("---------------------------------------------------------------------------\n")
	novel_keys = list(dict.fromkeys([a for b in novel_keys for a in b]))
	if novel_keys and preferences.comp_software == "gaussian":
		print("The following keys were not recognized:")
		print(novel_keys)
		print("---------------------------------------------------------------------------\n")
	try:
		import raapbs
		raapbs.option()
	except ImportError:
		pass


from raachem.util.interface import UserInterface
#############################HOOK on ERRORS###############################
import traceback, sys
def show_exception_and_exit(exc_type, exc_value, tb):
	traceback.print_exception(exc_type, exc_value, tb)
	input("Press key to exit.")
	sys.exit(-1)
sys.excepthook = show_exception_and_exit
##########################################################################
UserInterface().render()




import os, shutil, time, itertools, configparser

def read_item(file_name=None, promp=False, extension=None, cf=os.getcwd()):
	"""Reads an .xyz, .gjf, .com or .log item and returns a list of its contents ready for class instantiation"""
	if promp != False:
		if extension is None: extension = [".xyz"]
		print(promp)
		files = file_weeder(extension)
		if len(files) == 0: print("Sorry no such files in current directory!"); return []
		for idx,item in enumerate(files): print("{:<5}{:<25}".format(idx+1,item))
		while True:
			try: file_name = files[int(input())-1]; break
			except: print("Invalid input!")
	with open(os.path.join(cf,file_name),"r") as in_file: in_content = in_file.read().splitlines()
	in_content.insert(0,file_name)
	return in_content

def file_weeder(ext_to_weed,cf=os.getcwd(), promp=True):
	"""Looks up files with the extensions provided in current directory"""
	if type(ext_to_weed) == str: ext_to_weed = [ext_to_weed]
	weeded_list = []
	for a,b in itertools.product(ext_to_weed,os.listdir(cf)):
		if a in b: weeded_list.append(b)
	if promp and len(weeded_list) == 0:
		print("No '{}' files found in current directory!".format("' or '".join(ext_to_weed)))
		print(cf)
		return []
	return sorted(weeded_list)

def is_str_float(i):
	"""Check if a string can be converted into a float"""
	try: float(i); return True
	except ValueError: return False

def mv_up_folder():
	"""Move files with a chossen extension up a folder"""
	while True:
		extensions = [[0, "return"], ["1", ".log"], ["2", ".gjf"],["2", ".com"],["3", ".xyz"]]
		print("Select a file extension to move up a folder")
		for idx, ext in enumerate(extensions):
			if idx == 0: print("0   - Cancel request")
			else: print("{:<3} - {:>10}".format(ext[0], ext[1]))
		extension = {str(i[0]): i[1] for i in extensions}.get(input(), None)
		print(extension)
		if extension == "return": return
		if extension != None: break
		print("Invalid Option")
	folders = [x[0] for x in os.walk(os.getcwd())]
	for folder in folders[1:]:
		for file in file_weeder(extension, folder, promp=False):
			if os.path.isfile(os.path.join(folders[0], file)):
				print("Filename: {}\nFrom: {}\nAlready exists in uper directory:\n{}".format(file, folder, folders[0]))
				break
			shutil.copy2(os.path.join(folder, file), os.path.join(folders[0], file))
def sel_files(weeded_list):
	print("Choose the files you want to operate on:")
	print("(Multiple files can be separated by a space; 'a' for all files)")
	print("{:>3}{:>30}".format("0", "None"))
	for i,a in enumerate(weeded_list): print("{:>3}{:>30}".format(str(i+1),a))
	while True:
		option = input().lower().split()
		if all(b in ["a","0",*[str(a) for a in range(len(weeded_list)+1)]] for b in option): break
	if "0" in option: return False
	else: return weeded_list if "a" in option else [weeded_list[int(i)-1] for i in option]

def timeit(method):
	def timed(*args, **kw):
		ts = time.time()
		result = method(*args, **kw)
		te = time.time()
		print('{}:{:.2f} ms'.format(method.__name__, (te - ts) * 1000))
		return result
	return timed

class Var:
	conf_dir = os.path.dirname(__file__)
	conf_file = os.path.join(conf_dir, "user.init")

	def __init__(self,conf_file=conf_file):
		self.std_config = configparser.ConfigParser()
		self.std_config["SERVERS"] = {"heimdall_user": "heimdall_user",
							 "heimdall_mail": "heimdall_mail",
							 "heimdall_notification": "False",
							 "aguia_user": "aguia_user",
							 "athene_user": "athene_user",
							 "sub_s_name": "sub_s_name"}
		self.std_config["BEHAVIOUR"] = {"heavy_atom": "36",
							   "gjf_overwrite": "False",
							   "folder_op": "True",
							   "gauss_ext": ".com",
							   "comp_software": "gaussian"}
		self.std_config["ANALYSIS"] = {"options": "3 4 5 6 7"}
		if not os.path.isfile(conf_file):
			with open(conf_file, "w") as configfile:
				self.std_config.write(configfile)
			self.config = self.std_config
		else:
			self.config = configparser.ConfigParser()
			self.config.read(conf_file)

		def pick(args,get_type,valid_keys={},default=None,config=self.config,std_config=self.std_config):
			try:
				if get_type == "str":
					result = config.get(*args)
				elif get_type == "bool":
					result = config.getboolean(*args)
				elif get_type == "int":
					result = config.getint(*args)
			except:
				if get_type == "str":
					result = std_config.get(*args)
				elif get_type == "bool":
					result = std_config.getboolean(*args)
				elif get_type == "int":
					result = std_config.getint(*args)
			finally:
				if valid_keys:
					result = valid_keys.get(result,default)
				return result

		self.heimdall_user = pick(("SERVERS","heimdall_user"),"str")
		self.heimdall_mail = pick(("SERVERS","heimdall_mail"),"str")
		self.heimdall_notification = pick(("SERVERS","heimdall_notification"),"bool")
		self.aguia_user = pick(("SERVERS","aguia_user"),"str")
		self.athene_user = pick(("SERVERS","athene_user"),"str")
		self.sub_s_name = pick(("SERVERS","sub_s_name"),"str")
		keys={n:n for n in range(1,119)}
		self.heavy_atom = pick(("BEHAVIOUR","heavy_atom"),"int",valid_keys=keys,default=36)
		self.gjf_overwrite = pick(("BEHAVIOUR","gjf_overwrite"),"bool")
		self.folder_op = pick(("BEHAVIOUR","folder_op"),"bool")
		keys = {".gjf":".gjf","gjf":".gjf",".com":".com","com":".com"}
		self.gauss_ext = pick(("BEHAVIOUR","gauss_ext"),"str",valid_keys=keys,default=".com")
		keys = {"gaussian":"gaussian","g09":"gaussian","g16":"gaussian","orca":"orca"}
		self.comp_software = pick(("BEHAVIOUR","comp_software"),"str",valid_keys=keys,default="gaussian")
		self.e_options = [
			["Blank column","Blank"],
			["Hartree to kcal/mol conversion factor (627.5)", "Eh to kcal/mol"],
			["Hartree to kJ/mol conversion factor (2625.5)", "Eh to kJ/mol"],
			["Filename","Filename"],
			["Hyperlink to corresponding folder", "Folder"],
			["Hyperlink to Filename.xyz",".xyz"],
			["Hyperlink to Filename{}".format(self.gauss_ext),self.gauss_ext],
			["Route section read from Filename{}".format(self.gauss_ext), self.gauss_ext + "_#"],
			["Hyperlink to Filename.log",".log"],
			["Route section read from Filename.log",".log_#"],
			["Energy from last SCF cycle", "E0"],
			["Number of imaginary frequencies found on Filename.log","iFreq"],
			["Zero-point correction","E_ZPE"],
			["Thermal correction to Energy","E_tot"],
			["Thermal correction to Enthalpy","H_corr"],
			["Thermal correction to Gibbs Free Energy","G_corr"],
			["Sum of electronic and zero-point Energies","E0+E_ZPE"],
			["Sum of electronic and thermal Energies","E0+E_tot"],
			["Sum of electronic and thermal Enthalpies","E0+H_corr"],
			["Sum of electronic and thermal Free Energies","E0+G_corr"],
			["Filename.log gaussian normal termination status","Done?"],
			["Error messages found on Filename.log","Error"],
			["HOMO from Filename.log","HOMO"],
			["LUMO from Filename.log","LUMO"],
			["HOMO-LUMO gap from Filename.log","HOMO-LUMO"],
			["Charge from Filename.log","Charge"],
			["Starting multiplicity from Filename.log","Mult"],
			["Number of 'SCF Done:' keywords found","n_SCF"],
			["Number of atoms on Filename.log","n_atoms"],
			["Filename.log calculation type (This may be unreliable)","TYP"],
			["Filename.log calculation type consistency with iFreq","Needs refinement?"],
			["Filename.log last spin densities before anihilation", "S**2 BA"],
			["Filename.log last spin densities after anihilation", "S**2 After"],
			["Filename.log last geometry", "LG"],
			["Filename.log last Muliken charge and spin density","MulkSpinDens"],
			["Filename.log last Internal coordinates", "LastIntCoord"],
			["Filename.log last Muliken charge", "MulkCharges"],
			["Filename.log last ESP charge", "ESPCharges"],
			["Filename.log last population analysis", "POPAnalysis"],
			["Filename.log last NPA analysis", "NPAAnalysis"]

		]
		self.valid_e_keys = [str(n) for n in range(len(self.e_options))]
		self.e_analysis_opt = [int(a) for a in pick(("ANALYSIS","options"),"str").split() if a in self.valid_e_keys]
		#self.e_analysis_opt = [int(a) for a in self.valid_e_keys]
		global preferences
		preferences = self
	def set_variables(self):
		print("Please type 'chave' and press enter")
		if input().strip().lower() != "chave": return
		try: import raapbs;	pbs = True
		except ImportError:	pbs = False
		options = [
			["To return", None],
			["Set program BEHAVIOUR options", self.set_behaviour],
			["Set preferences related to the creation of xls files via energy analysis function",self.xls_creation]
		]
		if pbs:	options.insert(1,["Set SERVER related options", self.set_server])
		while True:
			option = self.get_option(options)
			if option == 0: return
			else: options[option][-1]()
	def get_option(self,options):
		print("Which variables do you want to set?")
		for i, v in enumerate(options):
			if v[1] is None: print("{} - {}".format(i, v[0]))
			else: print("{} - {} {}".format(i, v[0], "" if callable(v[1]) else getattr(self, v[1])))
		while True:
			option = input().strip()
			if option in [str(n) for n in range(len(options))]:	break
			else: print("Invalid Input!")
		return int(option)
	def set_server(self,conf_file=conf_file):
		options = [
			["To return", None],
			["HEIMDALL USER:", "heimdall_user"],
			["HEIMDALL MAIL:", "heimdall_mail"],
			["HEIMDALL SEND EMAIL (true/false):", "heimdall_notification"],
			["AGUIA USER:", "aguia_user"],
			["ATHENE USER:", "athene_user"],
			["SUBMISSION SCRIPT NAME:", "sub_s_name"]
		]
		while True:
			option = self.get_option(options)
			if option == 0: return
			else:
				variable = input("Enter variable '{}' value: ".format(options[option][-1]))
				self.config["SERVERS"][options[option][-1]]=variable.strip()
				with open(conf_file, "w") as configfile: self.config.write(configfile)
				self.__init__()
	def set_behaviour(self,conf_file=conf_file):
		options = [
			["To return", None],
			["ECP IS ADVISED FOR ELEMENTS LARGER THAN:", "heavy_atom"],
			["OVERWRITE .GJF FILES WITH NO PROMP (true/false):", "gjf_overwrite"],
			["AUTO OPERATE ON ALL FILES IN THE CWD (true/false):", "folder_op"],
			["GAUSSIAN INPUT FILE EXTENSION ('.gjf' or '.com'):", "gauss_ext"],
			["COMPUTATIONAL CHEMISTRY SOFTWARE (orca/gaussian):", "comp_software"]
		]
		while True:
			option = self.get_option(options)
			if option == 0: return
			else:
				variable = input("Enter variable '{}' value: ".format(options[option][-1]))
				self.config["BEHAVIOUR"][options[option][-1]]=variable.strip()
				with open(conf_file, "w") as configfile: self.config.write(configfile)
				self.__init__()
	def xls_creation(self,conf_file=conf_file):
		while True:
			print("=============================CURRENTLY INCLUDED COLUMNS=============================")
			correspondence = {}
			for i,a in enumerate(self.e_analysis_opt,start=1):
				print("{:>3} {:<59} {:>20}".format(i,*self.e_options[a]))
				correspondence.update({str(i):a})
			unused = [a for a in self.valid_e_keys if int(a) not in self.e_analysis_opt]
			start = len(correspondence)+1
			if unused:
				print("============================CURRENTLY UNINCLUDED COLUMNS============================")
			for i,a in enumerate(unused,start=start):
				print("{:>3} {:<59} {:>20}".format(i,*self.e_options[int(a)]))
				correspondence.update({str(i):int(a)})
			print("====================================================================================")
			print()
			print("xls analysis columns can be added with '+N' or removed with '-N':")
			print("Multiple operations can be separated with a space character. Enter '0' to go back")
			option = input().split()
			if "0" in option: return
			else:
				add_keys = [correspondence[a[1:]] for a in option if a[0] == "+" and a[1:] in correspondence.keys()]
				rem_keys = [correspondence[a[1:]] for a in option if a[0] == "-" and a[1:] in correspondence.keys()]
				all_keys = [a for a in self.e_analysis_opt if a not in rem_keys]
				all_keys.extend([a for a in add_keys if a not in rem_keys])
				self.config["ANALYSIS"]["options"] = " ".join([str(a) for a in all_keys])
				with open(conf_file, "w") as configfile:
					self.config.write(configfile)
				self.__init__()
preferences = Var()





from raachem.file_class.xyz import XyzFile
from raachem.util.constants import elements
from raachem.util.gen_purp import is_str_float
import functools, itertools, math

class LogFile:
	calc_types = ["TS","Red","IRC","Opt","SP"]
	def __init__(self,file_content):
		self.list = file_content
		self.lenght = len(self.list)
		self.name = self.list[0].strip()
		self.empty_line_idxs = []
		self.charge_mult = None
		self.input_geom_idx = None
		self.start_xyz_idxs = []
		self.end_resume_idxs = []
		self.start_resume_idxs = []
		self.linked_job_idxs = []
		self.multi_dash_idxs =[]
		self.scf_done = []
		####.thermal = ["ZPC","TCE","TCH","TCG","SZPE","STE","STH","STG"]
		self.thermal = [None, None, None, None, None , None, None, None]
		self.oc_orb_energies = []
		self.uno_orb_energies = []
		self.hash_line_idxs = []
		self.norm_term_idxs = []
		self.errors = []
		self.irc_points = []
		self.scan_points = []
		self.opt_points = []
		self.force_const_mat = []
		self.distance_matrix = []
		self.s_squared = []
		self.muliken_spin_densities_idxs = []
		self.muliken_charge_idxs = []
		self.chelpg_charge_idxs = []
		self.pop_analysis_idxs = []
		self.npa_start_idxs = []
		self.npa_end_idxs = []
		for i,a in enumerate(a.strip() for a in self.list):
			# i = index
			# a = line.strip()
			# b = line.split()
			# c = len(b)
			if a == "":                                                         self.empty_line_idxs.append(i); continue
			if a[-1] == "@":                                                    self.end_resume_idxs.append(i); continue
			elif a[0] == "1":
				if a.startswith(r"1\1"):                                      self.start_resume_idxs.append(i); continue
			if a[0].isdigit() or a[0].islower():                                                                continue
			elif a[0] == "-":
				if a.startswith("------"):                                      self.multi_dash_idxs.append(i); continue
			elif a[0] == "!":
				b = a.split(); c = len(b)
				if c == 4:
					condition_a = all(x in y for x,y in zip(b,("!",["Optimized","Non-Optimized"],"Parameters","!")))
					if condition_a:                                          self.scan_points.append([i,b[1]]); continue
			elif a[0] == "A":
				text_a = "Alpha  occ. eigenvalues --"
				text_b = "Alpha virt. eigenvalues --"
				text_c = "Atom  No          Natural Electron Configuration"
				if a.startswith(text_a):
					text = a.replace(text_a,"").replace("-"," -").split();   self.oc_orb_energies.extend(text); continue
				elif a.startswith(text_b):
					text = a.replace(text_b,"").split();                    self.uno_orb_energies.extend(text); continue
				elif a.split() == text_c.split():
					pass;                                                          self.npa_end_idxs.append(i); continue

			elif a[0] == "C":
				b = a.split(); c = len(b)
				if all((a.startswith("Charge"),self.charge_mult is None, c == 6)):
					pattern = ("Charge", "=", "Multiplicity", "=")
					if all(x == b[n] for x,n in zip(pattern,(0,1,3,4))):
						self.input_geom_idx = i;                                    self.charge_mult = b[2::3]; continue
			elif a[0] == "D":
				if a.startswith("Distance matrix (angstroms):"):   				self.distance_matrix.append(i); continue
			elif a[0] == "E":
				if a.startswith("Error"):                                                self.errors.append(i); continue
				elif a.startswith("ESP charges:"):                           self.chelpg_charge_idxs.append(i);continue
			elif a[0] == "F":
				if a.startswith("Full mass-weighted force constant matrix:"):   self.force_const_mat.append(i); continue
			elif a[0] == "I":
				if a == "Input orientation:":                                self.start_xyz_idxs.append(i + 5); continue
			elif a[0] == "L":
				if a.startswith("Link1:"):                                      self.linked_job_idxs.append(i); continue
			elif a[0] == "N":
				if a.startswith("Normal termination of Gaussian"):               self.norm_term_idxs.append(i); continue
			elif a[0] == "M":
				if a.startswith("Mulliken charges and spin densities:"):
					pass;                                           self.muliken_spin_densities_idxs.append(i); continue
				elif a.startswith("Mulliken charges:"):
					pass;                                                    self.muliken_charge_idxs.append(i);continue
				elif a.startswith("Molecular Orbital Coefficients:"):
					pass;                                                      self.pop_analysis_idxs.append(i);continue
			elif a[0] == "P":
				b = a.split(); c = len(b)
				if c != 6 or any(x != b[n] for x,n in zip(["Point","Number:","Path","Number:"],[0,1,3,4])):     continue
				if any(not b[n].isnumeric() for n in [2, 5]):                                                   continue
				else:                                                  self.irc_points.append([i, b[5], b[2]]); continue
			elif a[0] == "S":
				b = a.split(); c = len(b)
				if a == "Standard orientation:":                             self.start_xyz_idxs.append(i + 5); continue
				elif a.startswith("SCF Done:") and c > 5:                       self.scf_done.append([i,b[4]]); continue
				elif a.startswith("S**2 before annihil"):self.s_squared.append([i,b[3].replace(",",""),b[-1]]); continue
				elif a.startswith("Sum of electronic and zero-point Energies="):       self.thermal[4] = b[-1]; continue
				elif a.startswith("Sum of electronic and thermal Energies="):          self.thermal[5] = b[-1]; continue
				elif a.startswith("Sum of electronic and thermal Enthalpies="):        self.thermal[6] = b[-1]; continue
				elif a.startswith("Sum of electronic and thermal Free Energies="):     self.thermal[7] = b[-1]; continue
				elif a.startswith("Step") and c == 9:
					x = ["Step", "number", "out", "of", "a", "maximum", "of"]
					y = [0, 1, 3, 4, 5, 6, 7]
					z = all(b[n].isnumeric() for n in [2, 8])
					if all(d == b[n] for d,n in zip(x,y)) and z:                     self.opt_points.append(i); continue
			elif a[0] == "T":
				b = a.split()
				if a.startswith("Thermal correction to Energy="):                      self.thermal[1] = b[-1]; continue
				elif a.startswith("Thermal correction to Enthalpy="):                  self.thermal[2] = b[-1]; continue
				elif a.startswith("Thermal correction to Gibbs Free Energy="):         self.thermal[3] = b[-1]; continue
			elif a[0] == "Z":
				b = a.split()
				if a.startswith("Zero-point correction="):                             self.thermal[0] = b[-2]; continue
			elif a[0] == "#":                                                    self.hash_line_idxs.append(i); continue
			elif a[0] == "*":
				if a.replace("*","").startswith("Gaussian NBO Version 3.1"):
					pass;                                                         self.npa_start_idxs.append(i);continue
		#--------------------------------------------POST PROCESSING----------------------------------------------------
		x = None if self.start_xyz_idxs is None else [min(a for a in self.multi_dash_idxs if a > b) for b in self.start_xyz_idxs]
		self.end_xyz_idxs = x
		self.scan_end = [min(a for a in self.multi_dash_idxs if a > b[0]) for b in self.scan_points]
		try:
			x = [self.list[b:min(a for a in self.empty_line_idxs if a > b)] for b in self.force_const_mat]
			self.displ_block = x
		except Exception as e:
			print("Error while finding vibrational frequencies of log file")
			print(e)
			print(self.name)
			self.displ_block = []
		try:
			x = None if self.hash_line_idxs is None else min(a for a in self.multi_dash_idxs if a > self.hash_line_idxs[0])
			x = None if self.hash_line_idxs is None else "".join([a.lstrip() for a in self.list[self.hash_line_idxs[0]:x]])
			self.raw_route = " ".join(x.split())
		except IndexError as e:
			print("Error while finding route section of log file")
			print(e)
			print(self.name)
		try:
			self.oc_orb_energies = [float(b) for b in self.oc_orb_energies]
			self.uno_orb_energies = [float(a) for a in self.uno_orb_energies]
			self.homo = max(self.oc_orb_energies) if self.oc_orb_energies else None
			self.lumo = min(self.uno_orb_energies) if self.uno_orb_energies else None
			self.homo_lumo_gap = self.lumo - self.homo if not any([self.lumo is None, self.homo is None]) else None
		except ValueError as e:
			print("Error finding homo and lumo energies")
			print(e)
			print(self.name)
			self.homo, self.lumo, self.homo_lumo_gap = None, None, None
		if all([self.start_resume_idxs,self.end_resume_idxs]):
			x = ["".join([x.strip() for x in self.list[a:b]]).split("\\") for a,b in zip(self.start_resume_idxs,self.end_resume_idxs)]
			self.resumes = x
		else: self.resumes = None
		# --------------------------------------------------ASSURANCE---------------------------------------------------
		self.init_errors = []
		if self.charge_mult is None:
			self.init_errors.append("Charge and multiplicity could not be identified!")
		if len(self.start_resume_idxs) != len(self.end_resume_idxs):
			self.init_errors.append("Inconsistent resumes")
		if len(self.name.split()) != 1:
			self.init_errors.append("Name must not contain empty spaces or be empty")
		if not self.list[1].strip().startswith("Entering Gaussian System"):
			self.init_errors.append("Is this a Gaussian log file?")
		if not self.start_xyz_idxs is None:
			if len(self.start_xyz_idxs) != len(self.end_xyz_idxs):
				self.init_errors.append("Found an inconsistent number of geometries")
		if not any([self.homo is None, self.lumo is None]):
			if self.homo > self.lumo:
				self.init_errors.append("Lumo is lower than homo?")
		if self.init_errors:
			for a in self.init_errors: print(a)
			print("Errors above were found on file\n{}".format(self.name))
		#self._xyz_from_dist_matrix()
	@functools.lru_cache(maxsize=1)
	def loghelp(self):
		for a in vars(self):
			if a != "list":
				print(a.upper(),"--->",getattr(self,a))
	@functools.lru_cache(maxsize=1)
	def xyz_cord_block(self,start_idx,end_idx):
		data = [a.split() for a in self.list[start_idx:end_idx]]
		return [[elements[int(l[1])],*[l[i] for i in [3,4,5]]] for l in data]
	@functools.lru_cache(maxsize=1)
	def last_cord_block(self):
		if not all([self.xyz_cord_block, self.end_xyz_idxs]):
			if self.resumes:
				print("WARNING: Coordinates will be taken from the last job abstract:")
				print("lines {} - {} of file:".format(self.start_resume_idxs[-1],self.end_resume_idxs[-1]))
				print("{}".format(self.name))
				return self.last_log_abstract.xyz_object().cord_block()
			else: return None
		else:
			return self.xyz_cord_block(self.start_xyz_idxs[-1],self.end_xyz_idxs[-1])
	@functools.lru_cache(maxsize=1)
	def first_cord_block(self):
		if not all([self.start_xyz_idxs,self.end_xyz_idxs]):
			if self.input_geom_idx:
				coordinates = []
				for i,a in enumerate(self.list[self.input_geom_idx:]):
					if i > 5 and not coordinates: break
					a = a.split()
					if len(a) == 4:
						if a[0] in elements and all(is_str_float(a[n]) for n in [1, 2, 3]):
							coordinates.append(a)
						elif coordinates: break
					elif coordinates: break
				return coordinates
			else: return None
		else:
			return self.xyz_cord_block(self.start_xyz_idxs[0],self.end_xyz_idxs[0])
	@functools.lru_cache(maxsize=1)
	def _n_atoms(self):
		if self.last_cord_block():
			return len(self.last_cord_block())
		elif self.first_cord_block():
			return len(self.first_cord_block())

	def any_xyz_obj(self,a_idx,b_idx,title=" ",name=False):
		if name == False: name = self.name
		return XyzFile([name, self.n_atoms, title, *(" ".join(l) for l in self.xyz_cord_block(a_idx,b_idx))])
	@functools.lru_cache(maxsize=1)
	def last_xyz_obj(self):
		if self.last_cord_block():
			return XyzFile([self.name,self.n_atoms," ",*(" ".join(l) for l in self.last_cord_block())])
		else:
			return None

	@functools.lru_cache(maxsize=1)
	def first_xyz_obj(self):
		return XyzFile([self.name,self.n_atoms," ",*(" ".join(l) for l in self.first_cord_block())])

	@functools.lru_cache(maxsize=1)
	def low_e_xyz_obj(self):
		if self.calc_type == "SP": return None
		else:
			xyzs = {"TS":self.opt,"Red":self.scan_geoms,"IRC":self.irc,"Opt":self.opt}[self.calc_type]()
			if len(xyzs) == 0: return None
			else: return sorted(xyzs,key= lambda x: float(x.title()) if is_str_float(x.title()) else 1)[0]

	@functools.lru_cache(maxsize=1)
	def _calc_type(self):
		if self.raw_route:
			r_sect = [self.raw_route]
			for x in [None, "/", "(", ")", ",", "=", "%", ":"]:
				r_sect = [a for a in itertools.chain(*[i.split(x) for i in r_sect]) if len(a) > 1]
			r_sect = [a.lower() for a in r_sect]
			if any(a in r_sect for a in ("ts", "qst2","qst3")): return "TS"
			elif any(True for a in r_sect if a in ("modredundant", "readoptimize", "readfreeze")): return "Red"
			elif "irc" in r_sect: return "IRC"
			elif "opt" in r_sect: return "Opt"
			else: return "SP"
		else: return "No data"
	@functools.lru_cache(maxsize=1)
	def _normal_termin(self):
		return any(True if "Normal termination of Gaussian" in l else False for l in self.list[-5:])
	def _error_msg(self):
		error_idxs = [a for a in self.errors if a + 5 > self.lenght]
		for n in [-4,-3,-2,-1]:
			if self.list[n].strip().startswith("galloc:  could not allocate memory."):
				error_idxs.append(n)
		if error_idxs: return " | ".join([self.list[n] for n in error_idxs])
		else: return "No data"
	@functools.lru_cache(maxsize=1)
	def needs_ref(self):
		if self.calc_type == "Opt" and self.last_freq:
			if self.last_freq.n_ifreq() == "0": return "No"
			else: return "Yes"
		elif self.calc_type == "TS" and self.last_freq:
			if self.last_freq.n_ifreq() == "1": return "No"
			else: return "Yes"
		else: return "-"

	@functools.lru_cache(maxsize=1)
	def irc(self):
		if not all([self.start_xyz_idxs,self.end_xyz_idxs,self.irc_points,self.scf_done]): return []
		points = self.irc_points
		scf = [max(self.scf_done,key=lambda x: x[0] if x[0] < a[0] else 0)[1] for a in points]
		a_idx = [max(self.start_xyz_idxs,key=lambda x: x if x < a[0] else 0) for a in points]
		b_idx = [max(self.end_xyz_idxs,key=lambda x: x if x < a[0] else 0) for a in points]
		points = [[*d[1:],c,self.any_xyz_obj(a,b,title=c)] for a,b,c,d in zip(a_idx,b_idx,scf,points)]
		path_a = sorted([a for a in points if a[0] == "1"], key = lambda x: int(x[1]), reverse=True)
		path_b = [a for a in points if a[0] == "2"]
		return [a[3] for a in [*path_a,*path_b]]
	@functools.lru_cache(maxsize=1)
	def opt(self):
		if not all([self.start_xyz_idxs,self.end_xyz_idxs,self.opt_points,self.scf_done]): return []
		points = self.opt_points
		scf = [max(self.scf_done,key=lambda x: x[0] if x[0] < a else 0)[1] for a in points]
		a_idx = [max(self.start_xyz_idxs,key=lambda x: x if x < a else 0) for a in points]
		b_idx = [max(self.end_xyz_idxs,key=lambda x: x if x < a else 0) for a in points]
		return [self.any_xyz_obj(a,b,title=c) for a,b,c in zip(a_idx,b_idx,scf)]
	@functools.lru_cache(maxsize=1)
	def scan_geoms(self):
		if not all([self.start_xyz_idxs, self.end_xyz_idxs, self.scan_points, self.scf_done]): return []
		geoms = []
		all_points = self.scan_points
		points = [a for a in all_points if a[0] < self.start_xyz_idxs[-1] and a[0] < self.end_xyz_idxs[-1]]
		points_removed = len(all_points) - len(points)
		if points_removed != 0:
			print(f"WARNING: {points_removed} Scan  points have been removed due to inconsistent number o geometries found")
		start_idx = [min(i for i in self.start_xyz_idxs if i > b[0])  for b in points]
		end_idx = [min(i for i in self.end_xyz_idxs if i > b[0]) for b in points]
		scf_idx = [max(i for i in self.scf_done if i[0] < b[0]) for b in points]
		for i,(a,b,c,d) in enumerate(zip(start_idx,end_idx,scf_idx,points)):
			name = self.name.replace(".log","_" + str(i+1)+".xyz")
			if d[1] == "Optimized": print("Optimized geometry found at line {}!".format(d[0]))
			elif d[1] == "Non-Optimized": print("Non-Optimized1 geometry found at line {}!".format(d[0]))
			geoms.append(self.any_xyz_obj(a,b,title=str(c[1]), name=name))
		if len(geoms) == 0:
			print("No Optimized geometries found for {} file".format(self.name()))
		return geoms
	@functools.lru_cache(maxsize=1)
	def _last_freq(self):
		return LogFreq(self.displ_block[-1]) if self.displ_block else False
	@functools.lru_cache(maxsize=1)
	def _last_log_abstract(self):
		return LogAbstract(self.resumes[-1]) if self.resumes is not None else False
	def _xyz_from_dist_matrix(self):
		end_idx = lambda x: next(i for i,a in enumerate(self.list[x+1:],start=x+1) if not a.split()[0].isdigit())
		return [DistMatrix(self.list[a+1:end_idx(a)]) for a in self.distance_matrix]
	def _last_muliken_spin_density(self):
		if self.muliken_spin_densities_idxs:
			end_idx = lambda x: next(i for i, a in enumerate(self.list[x + 1:], start=x + 1) if not a.split()[0].isdigit())
			return "\n".join(self.list[self.muliken_spin_densities_idxs[-1]:end_idx(self.muliken_spin_densities_idxs[-1]+1)])
		else:
			return "-"
	def _last_internal_coord(self):
		if self.scan_points:
			end_idx = lambda x: next(i for i, a in enumerate(self.list[x + 1:], start=x + 1) if not a.strip().startswith("!"))
			return "\n".join(self.list[self.scan_points[-1][0]-1:end_idx(self.scan_points[-1][0]+5)+1])
		else:
			return "-"
	def _last_muliken_charges(self):
		if self.muliken_charge_idxs:
			end_idx = lambda x: next(i for i, a in enumerate(self.list[x + 1:], start=x + 1) if not a.split()[0].isdigit())
			return "\n".join(self.list[self.muliken_charge_idxs[-1]:end_idx(self.muliken_charge_idxs[-1]+1)])
		else:
			return "-"
	def _last_chelpg_charges(self):
		if self.chelpg_charge_idxs:
			end_idx = lambda x: next(i for i, a in enumerate(self.list[x + 1:], start=x + 1) if not a.split()[0].isdigit())
			return "\n".join(self.list[self.chelpg_charge_idxs[-1]:end_idx(self.chelpg_charge_idxs[-1] + 1)])
		else:
			return "-"
	def _pop_analysis(self):
		if self.pop_analysis_idxs and self.muliken_charge_idxs:
			return "\n".join(self.list[self.pop_analysis_idxs[-1]:self.muliken_charge_idxs[-1]])
		else:
			return "-"
	def _npa_analysis(self):
		if self.npa_start_idxs and self.npa_end_idxs:
			if len(self.npa_start_idxs) > 1:
				end_idx = lambda x: next(i for i, a in enumerate(self.list[x + 1:], start=x + 1) if a.strip() == "")
				return "\n".join(self.list[self.npa_start_idxs[-2]:end_idx(self.npa_end_idxs[-1])])
			else:
				return "-"
		else:
			return "-"
	n_atoms = property(_n_atoms)
	normal_termin = property(_normal_termin)
	calc_type = property(_calc_type)
	error_msg = property(_error_msg)
	last_log_abstract = property(_last_log_abstract)
	last_freq = property(_last_freq)
	xyz_from_dist_matrix = property(_xyz_from_dist_matrix)
	last_muliken_spin_density = property(_last_muliken_spin_density)
	last_internal_coord = property(_last_internal_coord)
	last_muliken_charges = property(_last_muliken_charges)
	last_chelpg_charges = property(_last_chelpg_charges)
	pop_analysis = property(_pop_analysis)
	npa_analysis = property(_npa_analysis)

class LogAbstract:
	def __init__(self,content):
		assert type(content) is list
		self.list = content
		self.version = None
		self.dipole = None
		self.img_freq = None
		self.hash_line = None
		for i,a in enumerate(self.list):
			a = a.lstrip()
			if a.lstrip == "": print("Empty!");continue
			elif a.startswith("Version="): self.version = a.replace("Version=","")
			elif a.startswith("#"): self.hash_line = i
			elif a.startswith("NImag="): self.img_freq = a.replace("NImag=0","")
			elif a.startswith("DipoleDeriv="): self.img_freq = a.replace("DipoleDeriv=","")
			else: continue
	def __str__(self):
		return "\n".join(self.list)
	def read_strucure(self):
		charge_mult = None
		title = None
		coordinates = []
		for i,a in enumerate(self.list[self.hash_line:]):
			if i > 5 and not coordinates: break
			a = a.split(",")
			if len(a) == 2 and not coordinates:	charge_mult = a; continue
			if len(a) == 4:
				if a[0] in elements and all(is_str_float(a[n]) for n in [1,2,3]):
					coordinates.append("   ".join(a))
				elif coordinates: break
			elif coordinates: break
		return charge_mult, XyzFile([self.list[0],str(len(coordinates)),title,*coordinates])
	def charge_mult(self):
		return self.read_strucure()[0]
	def xyz_object(self):
		return self.read_strucure()[1]

class LogFreq:
	def __init__(self, content):
		assert type(content) is list
		self.list = content
		self.rows = []
		for i,a in enumerate(self.list):
			if a.lstrip().startswith("Frequencies --"):
				try:
					assert self.list[i + 1].lstrip().startswith("Red. masses --")
					assert self.list[i + 2].lstrip().startswith("Frc consts  --")
					assert self.list[i + 3].lstrip().startswith("IR Inten    --")
					self.rows.append(i-2)
				except AssertionError:
					continue
		if not self.rows:
			self.n_atoms = 1
			self.block = []
		else:
			self.n_atoms = len(self.list) - self.rows[-1] - 7
			self.block = self.list[self.rows[0]:]
	def __str__(self):
		return "\n".join(self.list)
	@functools.lru_cache(maxsize=1)
	def frequencies(self):
		return list(itertools.chain(*[i.split()[2:] for i in self.block[2::self.n_atoms+7]]))
	@functools.lru_cache(maxsize=1)
	def ir_intensities(self):
		return list(itertools.chain(*[i.split()[3:] for i in self.block[5::self.n_atoms+7]]))
	@functools.lru_cache(maxsize=1)
	def displ_for_freq_idx(self,freq_idx):
		displ = []
		for num in range(self.n_atoms):
			displ.append(list(itertools.chain(*[i.split()[2:] for i in self.block[7+num::self.n_atoms+7]])))
		displ_for_freq_str = [a[freq_idx*3:freq_idx*3+3] for a in displ]
		displ_for_freq_float = [[float(i) for i in b] for b in displ_for_freq_str]
		return displ_for_freq_float
	@functools.lru_cache(maxsize=1)
	def n_ifreq(self):
		return str(len([b for b in self.frequencies() if float(b) < 0])) if self.frequencies() else "No data"
	def ir_spectra(self,threshold = 20):
		pairs = []
		for a,b in zip(self.frequencies(), self.ir_intensities()):
			if is_str_float(a) and is_str_float(b):
				pairs.append([float(a),float(b)])
		for a,b in zip(sorted(pairs,key=lambda x: x[1], reverse=True), range(threshold)):
			print("{:>10.1f}{:>10.1f}".format(float(a[0]),float(a[1])))
		print("---------")
class DistMatrix:
	def __init__(self,text):
		self.element = {}
		for a in [b.split() for b in text]:
			idx = "".join(a[0:2])
			if len(a) > 2 and a[1] in elements:
				if idx in self.element:
					self.element[idx].extend([float(c) for c in a[2:]])
					continue
				else:
					self.element[idx] = [float(c) for c in a[2:]]
		for a in self.element:
			print(a,self.element[a])
		self.dist_matrix = sorted(self.element.values(),key=lambda x: len(x))
		self.elem_vector = sorted(self.element.keys(), key=lambda x: len(self.element[x]))
		self.xyz_ent_a = []
		self.xyz_ent_b = []
		for i,a in enumerate(self.dist_matrix):
			if i == 0:
				self.xyz_ent_a.append([0, 0, 0])
				self.xyz_ent_b.append([0, 0, 0])
			if i == 1:
				self.xyz_ent_a.append([a[0], 0, 0])
				self.xyz_ent_b.append([a[0], 0, 0])
			if i == 2:
				x = (self.dist_matrix[i-1][0]**2+a[0]**2-a[1]**2)/(2*self.dist_matrix[i-1][0])
				y = math.sqrt(a[1]**2-x**2)
				self.xyz_ent_a.append([x, y, 0])
				self.xyz_ent_b.append([x, y, 0])
			if i > 2:
				x = (self.dist_matrix[i-1][0]**2+a[0]**2-a[1]**2)/(2*self.dist_matrix[i-1][0])
				y = math.sqrt(a[1]**2-x**2)
				#z =
				self.xyz_ent_a.append([x, y, 0])
				self.xyz_ent_b.append([x, y, 0])




from raachem.util.constants import elements
from raachem.file_class.xyz import XyzFile
import functools, re
class GjfFile:
	pattern = re.compile(r"[0-9][0-9][Gg]\+")
	def __init__(self,file_content):
		self.list = file_content
		self.list_l = [a.split() for a in file_content]
		self.str_l = [a.replace(" ", "") for a in self.list]
		self.return_print = "\n".join(self.list[1:])
		#########################
		#########################
		self.empty_line_idxs = [i for i,a in enumerate(self.list) if a.split() == []]
		self.asterisk_line_idxs = [idx for idx,line in enumerate(self.list) if line.split() == ["****"]]
		self.link_one_idxs = [i for i,l in enumerate(self.list) if "--link1--" in l.lower()]

	@functools.lru_cache(maxsize=1)
	def name(self):
		if len(self.list[0]) == 0: raise Exception(".gjf or .com object has no name")
		assert type(self.list[0]) is str, "Name must be string"
		return self.list[0]
	@functools.lru_cache(maxsize=1)
	def charge(self):
		return int(self.list[self.c_m_idx()].split()[0])
	@functools.lru_cache(maxsize=1)
	def multiplicity(self):
		return int(self.list[self.c_m_idx()].split()[1])
	@functools.lru_cache(maxsize=1)
	def n_electrons(self):
		return sum(elements.index(e) for e in self.all_elements()) - self.charge()
	@functools.lru_cache(maxsize=1)
	def n_atoms(self):
		return len(self.all_elements())
	@functools.lru_cache(maxsize=1)
	def all_elements(self):
		return [line[0] for line in self.cord_block()]
	@functools.lru_cache(maxsize=1)
	def elements(self):
		return list(dict.fromkeys(self.all_elements()))
	@functools.lru_cache(maxsize=1)
	def c_m_validate(self):
		return not self.n_electrons()%2 == self.multiplicity()%2
	@functools.lru_cache(maxsize=1)
	def c_m_validate_txt(self):
		return "Yes" if self.c_m_validate() else "--NO!--"
	@functools.lru_cache(maxsize=1)
	def n_proc(self):
		for line in self.list:
			line = line.lower().replace(" ","")
			if "%nprocshared=" in line:	return int(line.replace("%nprocshared=",""))
			elif "%nproc=" in line:	return int(line.replace("%nproc=",""))
	@functools.lru_cache(maxsize=1)
	def cord_block(self):
		coordinates = []
		for line in self.list_l[self.c_m_idx()+1:]:
			if len(line) == 0: break
			if len(line) != 4: continue
			if line[0] in elements:	coordinates.append(line)
			else: coordinates.append([elements[int(line[0])],*line[0:]])
		return coordinates
	@functools.lru_cache(maxsize=1)
	def route_text(self):
		flatten = lambda l: [item for sublist in l for item in sublist]
		try: return " ".join(flatten([a.split() for a in self.list[self.route_idx():self.title_idx()]]))
		except: return "No data"
	@functools.lru_cache(maxsize=1)
	def c_m_idx(self):
		if len(self.list[self.title_idx()+2].split()) < 2:
			raise Exception("Did you provide charge and multiplicity data at line {} of file {}?".format(self.title_idx()+1,self.name()))
		return self.title_idx()+2
	@functools.lru_cache(maxsize=1)
	def end_cord_idx(self):
		for idx,line in enumerate(self.list):
			if idx < self.c_m_idx(): continue
			if line.split() == []: return idx+1
	#########################
	#########################
	@functools.lru_cache(maxsize=1)
	def route_idx(self):
		for idx,line in enumerate(self.list):
			if line.strip().startswith("#"):return idx
		raise Exception("A route section (#) should be specified for .gjf or .com files")
	@functools.lru_cache(maxsize=1)
	def title_idx(self):
		for idx,line in enumerate(self.list):
			if idx > self.route_idx() and line.split() == []: return idx+1
	@functools.lru_cache(maxsize=1)
	def gen_basis(self):
		return any(i in self.route_text().lower() for i in ["/gen", "gen ","genecp"])
	@functools.lru_cache(maxsize=1)
	def declared_basis_lines(self):
		if not self.gen_basis(): return None
		idxs = [i+1 for idx,i in enumerate(self.asterisk_line_idxs) if i < self.asterisk_line_idxs[-1]]
		idxs.insert(0,max(i+1 for i in self.empty_line_idxs if  i < self.asterisk_line_idxs[-1]))
		return idxs
	@functools.lru_cache(maxsize=1)
	def declared_basis(self):
		e_w_b = [self.list[i].split()[:-1] for i in self.declared_basis_lines()]
		return [j.capitalize() for i in e_w_b for j in i]
	@functools.lru_cache(maxsize=1)
	def basis_errors(self):
		if not self.gen_basis(): return []
		#errors
		zero_last = any(self.list[i].split()[-1] == "0" for i in self.declared_basis_lines())
		miss_basis = [a for a in self.elements() if a not in self.declared_basis()]
		surpl_basis = [a for a in self.declared_basis() if a not in self.elements()]
		rep_basis = list(dict.fromkeys([a for a in self.declared_basis() if self.declared_basis().count(a) > 1]))
		errors = []
		for i in [*[a+1 for a in self.declared_basis_lines()],self.route_idx()]:
			if GjfFile.pattern.search(self.list[i]):
				errors.append("Is the basis set specifications correct?".format(i))
				errors.append("{}".format(self.list[i]))
				errors.append("Shouldn't '+' appear before the letter 'G'?")
		#statements
		if not zero_last:errors.append("Missing zero at the end of basis set specification?")
		if miss_basis:errors.append("Missing basis for: {} ?".format(" ".join(miss_basis)))
		if surpl_basis:errors.append("Surplous basis for: {} ?".format(" ".join(surpl_basis)))
		if rep_basis:errors.append("Repeated basis for: {} ?".format(" ".join(rep_basis)))
		return errors
	@functools.lru_cache(maxsize=1)
	def gen_ecp(self):
		return any(i in self.route_text().lower() for i in ["pseudo", "genecp"])
	@functools.lru_cache(maxsize=1)
	def declared_ecp_lines(self):
		line_idx = []
		if not self.gen_ecp(): return None
		if self.gen_basis(): start_idx = self.declared_basis_lines()[-1] + 1
		else:start_idx = self.end_cord_idx()
		for idx,line in enumerate(self.list):
			if idx < start_idx: continue
			if len(line.split()) <= 1: continue
			if line.split()[-1] != "0": continue
			if all(True if a.capitalize() in elements else False for a in line.split()[:-1]): line_idx.append(idx)
		return line_idx
	@functools.lru_cache(maxsize=1)
	def declared_ecp(self):
		ecps = [self.list[i].split()[:-1] for i in self.declared_ecp_lines()]
		return [j.capitalize() for i in ecps for j in i]
	@functools.lru_cache(maxsize=1)
	def ecp_errors(self,heavy_e = 36):
		if not self.gen_ecp(): return []
		#errors
		zero_last = any(self.list[i].split()[-1] == "0" for i in self.declared_ecp_lines())
		miss_ecp = [a for a in self.elements() if a not in self.declared_ecp() and elements.index(a) > heavy_e]
		surpl_ecp = [a for a in self.declared_ecp() if a not in self.elements()]
		rep_ecp = list(dict.fromkeys([a for a in self.declared_ecp() if self.declared_ecp().count(a) > 1]))
		#statements
		errors = []
		if not zero_last:errors.append("Missing zero at the end of ecp set specification?")
		if miss_ecp:errors.append("Missing ecp for: {} ?".format(" ".join(miss_ecp)))
		if surpl_ecp:errors.append("Surplous ecp for: {} ?".format(" ".join(surpl_ecp)))
		if rep_ecp:errors.append("Repeated ecp for: {} ?".format(" ".join(rep_ecp)))
		return errors
	@functools.lru_cache(maxsize=1)
	def route_errors(self):
		errors = []
		keywords = self.route_text().lower().split()
		if len(keywords) > 1:
			if "nosymm" in 	keywords:
				if keywords[0] == "#t" or keywords[0:2] == ["#","t"]:
					errors.append("Combination of 'NoSymm' and '#T' might supress geometry output!")
		return errors


	@functools.lru_cache(maxsize=1)
	def mem(self):
		for line in self.list:
			line = line.lower().replace(" ","")
			if line.startswith("%mem=") and line.endswith("mb"): return int(line[5:-2])
			elif line.startswith("%mem=") and line.endswith("gb"): return 1000*int(line[5:-2])
		return None
	#########################
	#########################
	def replace_cord(self, xyz_obj):
		new = []
		for line in self.list[0:self.c_m_idx() + 1]: new.append(line)
		for line in xyz_obj.form_cord_block(): new.append(line)
		for line in self.list[self.end_cord_idx()-1:]: new.append(line)
		return GjfFile(new)
	def xyz_obj(self):
		return XyzFile([self.name(),self.n_atoms()," ",*[" ".join(a) for a in self.cord_block()]])


#!/usr/bin/env python
import os
current_folder=str(os.path.dirname(os.path.abspath(__file__)))
print("========================")
print("====Enter 0 to cancel===")
print("========================")
def evaluate_list(folder,level=0,scripts=[" "]):
	for file in os.listdir(folder):
		if os.path.isdir(os.path.join(folder,file)):
			#print("{0:}{1:}".format(" "*level, file))
			level += 1
			evaluate_list(os.path.join(folder,file),level,scripts)
		if file == "sub_s_name":
			scripts.append(os.path.join(folder,file))
			print("{0:<2}{1:}".format(len(scripts)-1," - " + scripts[-1]))
			os.system("chmod 755 {0:}".format(os.path.join(folder,file)))
	return	scripts
try:	
	filename = evaluate_list(current_folder)[int(input())]
	if not filename == " ":
		os.chdir(filename.replace("sub_s_name",""))
		os.system("{0:}".format(filename))
		#os.rename(filename,filename.replace("_sub","_subed"))
	print("Done!")

except IndexError:
	print("Not a valid entry!")
except NameError:
	print("Should be a number!")



import os, random, math
from raachem.util.gen_purp import read_item, file_weeder, preferences, sel_files
from raachem.file_class.log import LogFile
from raachem.file_class.xyz import XyzFile
from raachem.file_class.inp import InpFile
from raachem.file_class.gjf import GjfFile

def xyz_ent():
	weeded_list = file_weeder([".xyz"])
	weeded_list = weeded_list if preferences.folder_op else sel_files(weeded_list)
	if not weeded_list: return
	for i in weeded_list:
		xyz = XyzFile(read_item(i)).enantiomer()
		xyz.save_file()
def input_to_xyz():
	extension = preferences.gauss_ext if preferences.comp_software == "gaussian" else ".inp"
	weeded_list = file_weeder([extension])
	weeded_list = weeded_list if preferences.folder_op else sel_files(weeded_list)
	if not weeded_list: return
	for i in weeded_list:
		if preferences.comp_software == "orca":
			xyz = InpFile(read_item(i)).xyz_obj()
			xyz.save_file()
		elif preferences.comp_software == "gaussian":
			xyz = GjfFile(read_item(i)).xyz_obj()
			xyz.save_file()

def log_to_xyz():
	weeded_list = file_weeder([".log"])
	weeded_list = weeded_list if preferences.folder_op else sel_files(weeded_list)
	if not weeded_list: return
	while True:
		print("Which geometry do you want?")
		print("0 - Cancel")
		print("1 - Last")
		print("2 - Lowest energy")
		print("3 - First")
		geom = input()
		if geom == "0": return
		if geom in ["1","2","3"]: break
		else: print("Invalid option!")
	for i in weeded_list:
		try:
			if geom == "1":	LogFile(read_item(i)).last_xyz_obj().save_file()
			elif geom == "2": LogFile(read_item(i)).low_e_xyz_obj().save_file()
			elif geom == "3": LogFile(read_item(i)).first_xyz_obj().save_file()
		except Exception as e:
			print("Error on file: {}".format(i))
			print(e)

def log_to_xyz_scan():
	weeded_list = sel_files(file_weeder([".log"]))
	for item in weeded_list:
		log = LogFile(read_item(item))
		if log.calc_type.lower() == "red": xyzs = log.scan_geoms(); ext_name = "_scan_traject.xyz"
		elif log.calc_type.lower() == "irc": xyzs = log.irc(); ext_name = "_irc_traject.xyz"
		elif log.calc_type.lower() in ["opt","ts"]: xyzs = log.opt(); ext_name = "_opt_traject.xyz"
		else: print("The file is not a scan, opt or irc log file"); continue
		if len(xyzs) < 2:
			print("The file contains less than 2 steps")
			is_terse_irc = all([len(xyzs) < 2, log.calc_type == "IRC",
				log.raw_route.lower().startswith("#t") or log.raw_route.lower().startswith("# t ")])
			if is_terse_irc:
				print("WARNING! Error above may have been caused by the following:")
				print("IRC calculations with the #T keyword may not archive intermediate geometries")
				print("#N or #P keywords are recommended instead")
			continue
		max_e = max(float(i.title()) for i in xyzs)
		min_e = min(float(i.title()) for i in xyzs)
		c=[0,"percentage","kcal"]
		print("{:5}{:^90}{:>8}".format("Entry"," >---> Relative Energy >---> "," kcal"))
		for entry in [float(i.title()) for i in xyzs]:
			c[0]=c[0]+1
			c[1]= "|"*int(90*((float(entry)-min_e)/(max_e-min_e)) if max_e != min_e else 0)
			c[2]= (float(entry)-min_e)*627.5
			print("{:<5}{:<90}".format(c[0],c[1]),"{:>6.5f}".format(c[2]) if ext_name == "_opt_traject.xyz" else "{:>6.2f}".format(c[2]))
		file_name = item.replace(".log",ext_name)
		with open(file_name,mode="w",newline="\n") as file: file.write("\n".join([a for i in xyzs for a in i.return_print()]))

def log_freq_xyz():
	weeded_list = sel_files(file_weeder([".log"]))
	for i in weeded_list:
		log = LogFile(read_item(i))
		if not log.last_freq:
			print("No frequencies found in file:{}!".format(i))
			continue
		option = None
		while True:
			print("Analizing file: {}".format(i))
			print("Chose a frequency to split (0-To cancel/m-For more)")
			for idx,item in enumerate(log.last_freq.frequencies(),start=1):
				print("{:>5}:{:>15}".format(idx,item))
				if option == "m": continue
				if idx > 2: break
			option=input()
			if option == "m": continue
			elif option in [str(a+1) for a in range(len(log.last_freq.frequencies()))]: break
			elif option == "0": break
			else: print("Invalid option, try again!")
		if option == "0": continue
		print("Give a multiplier (recomended multiplier = 1) ")
		while True:
			try:
				mult=float(input().replace(",","."))
				break
			except ValueError:
				print("Not a valid number")
		left = log.last_xyz_obj().displace(-mult,log.last_freq.displ_for_freq_idx(int(option)-1))
		left.list[0] = i.replace(".log","_l.xyz")
		left.save_file()
		right = log.last_xyz_obj().displace(mult,log.last_freq.displ_for_freq_idx(int(option)-1))
		right.list[0] = i.replace(".log","_r.xyz")
		right.save_file()
		print("\nJob Done!\n")
	print("Finished !")

def superimpose_alg():
	while True:
		print("Chosse an option:")
		print("0 - To cancel")
		print("1 - For batch mode")
		print("2 - For single structure(In screen printing)")
		mode = input()
		if mode in [str(n) for n in range(3)]: break
	if mode == "0":
		return
	elif mode == "1":
		if os.path.exists(os.path.join(os.getcwd(), "rotated")):
			print("Rotated directory already exists in current directory!")
			print("Please remove it and try again.")
			return
		os.mkdir(os.path.join(os.getcwd(), "rotated"))

		xyz_1 = read_item(None, "Which item do you wish to compare to?", [".xyz"])
		if xyz_1: xyz_1 = XyzFile(xyz_1)
		else:return

		while True:
			num_atoms = input("compare first N atoms (0 for all):\n")
			if num_atoms in [str(a) for a in range(xyz_1.n_atoms())]:
				num_atoms = int(num_atoms)
				break
			else:
				print("Invalid input for {}".format(xyz_1.name()))
		for file in file_weeder(["xyz"]):
			if not file == xyz_1.name():
				xyz_2 = XyzFile(read_item(file))
				xyz_3 = xyz_2.superimpose(xyz_1, num_atoms, True)
				xyz_3.save_file(os.path.join(os.getcwd(), "rotated"))
		xyz_1.std_cord(num_atoms).save_file(os.path.join(os.getcwd(), "rotated"))
	elif mode == "2":
		xyz_1 = read_item(None, "Which item do you wish to compare to?", [".xyz"])
		if xyz_1: xyz_1 = XyzFile(xyz_1)
		else: return

		xyz_2 = read_item(None, "Which item do you wish to rotate?", [".xyz"])
		if xyz_2: xyz_2 = XyzFile(xyz_2)
		else:return

		while True:
			num_atoms = input("compare first N atoms (0 for all):\n")
			if num_atoms in [str(a) for a in range(min([xyz_1.n_atoms(),xyz_2.n_atoms()]))]:
				num_atoms = int(num_atoms)
				break
			else:
				print("Invalid input for either or both {} and {}".format(xyz_1.name(),xyz_2.name()))
		xyz_2.superimpose(xyz_1, num_atoms, True).print_file()
		xyz_1.std_cord(num_atoms).print_file()

def geodes_int():
	"""Requests input for solvation algorithm"""
	def rot_alg(xyz_1, xyz_2, angles, p_angles, geodes, save):
		"""Rotate xyz_2 by angles and p_angles then aproximate it to xyz_1 on geodes trajectories"""
		trajectory = []
		trajectory_mult = []
		xyz_old = xyz_2
		iteration = 0
		for axis in p_angles:
			xyz_2 = xyz_old.rotate(axis[1], axis[0])
			for angle in angles:
				xyz_2 = xyz_2.rotate(angle, "y")
				trajectory_other = []
				for dot in geodes:
					xyz_3 = aproximate(xyz_1, xyz_2, dot)
					xyz_4 = xyz_1 + xyz_3
					trajectory.append(xyz_4.return_print())
					trajectory_other.append(xyz_3.return_print())
					iteration += 1
					if save == "y":
						xyz_4.list[0] = "{}_{}.xyz".format(os.path.splitext(xyz_4.name())[0], str(iteration))
						xyz_4.save_file()
				trajectory_other.append(xyz_1.return_print())
				cord = [j for i in trajectory_other for j in i[2:]]
				trajectory_mult.append(str(len(cord)) + "\n\n")
				for line in cord:
					trajectory_mult.append(str(line) + "\n")
		# save suporposed trajectory
		file_path = os.path.join(os.getcwd(), "superposed.xyz")
		with open(file_path, "w") as file:
			for geom in trajectory_mult:
				file.write(str(geom))
		file_path = os.path.join(os.getcwd(), "trajectory.xyz")
		# save parwise trajectory
		with open(file_path, "w") as file:
			for xyz in trajectory:
				for line in xyz:
					file.write(str(line) + "\n")

	def aproximate(xyz_1, xyz_2, vertex):
		"""Aproximate xyz_2 to xyz_1 from trajectory vertex. Returns [xyz_1+xyz_2, xyz_2]"""

		def min_dist(a, b):
			return min(math.sqrt(sum(pow(c - d, 2) for c, d in zip(bb, aa))) for bb in b for aa in a)

		xyz_a = [[float(a) for a in i] for i in xyz_1.cord_strip()]
		xyz_b = [[float(a) + b * 50 for a, b in zip(i, vertex)] for i in xyz_2.cord_strip()]
		for num in range(5):
			step_size = 1 / 2 ** num
			while True:
				if min_dist(xyz_a, xyz_b) < 2:
					xyz_b = [[a + b * step_size for a, b in zip(i, vertex)] for i in xyz_b]
					break
				else:
					xyz_b = [[a - b * step_size for a, b in zip(i, vertex)] for i in xyz_b]
		cordinates = [" ".join([a, *[str(c) for c in b]]) for a, b in zip(xyz_2.all_elements(), xyz_b)]
		return XyzFile([xyz_2.name(), xyz_2.n_atoms(), " ", *cordinates])
	def start():
		angles, p_angles, geodes = None, None, None
		xyz_1 = read_item(None, "Choose the molecule you want to solvate?", ["xyz"])
		if xyz_1: xyz_1 = XyzFile(xyz_1).std_cord()
		else: return

		while True:
			print("Choose a trajectory englobing method")
			print("0 - To cancel")
			print("1 - Cubic gobling")
			print("2 - Icosahedron gobling")
			print("3 - Give a fulerene")
			usr_opt = input()
			if usr_opt == "0": return
			if usr_opt in ["1","2","3"]: break
		if usr_opt == "1":
			geodes = [[a,b,c] for a in [-1,1] for b in [-1,1] for c in [-1,1]]
		elif usr_opt == "2":
			k = (1 + math.sqrt(5)) / 2
			geodes = [[0, 1, k], [0, -1, k], [0, -1, -k],
					[0, 1, -k], [k, 0, 1], [-k, 0, 1],
					[-k, 0, -1],[k, 0, -1], [1, k, 0],
					[-1, k, 0], [-1, -k, 0], [1, -k, 0]]
		elif usr_opt == "3":
			geodes = read_item(None, "Chose a fulerene", ["xyz"])
			if geodes:
				geodes = XyzFile(geodes).std_cord().cord_strip()
				geodes = [[float(n) for n in i] for i in geodes]
			else:
				print("You need a to choose a trajectory engobling method")
				return

		xyz_2 = read_item(None, "Chose the solvent or counterion", ["xyz"])
		if xyz_2: xyz_2 = XyzFile(xyz_2).std_cord()
		else: return

		while True:
			print("Orientation algorithm:")
			print("0 - To cancel")
			print("1 - Espherical ion")
			print("2 - 12 Random cubic orientations")
			print("3 - 24 Sistematic cubic rotations")
			usr_opt = input()
			if usr_opt in [str(a) for a in range(4)]: break
		if usr_opt == "0":
			return
		elif usr_opt == "1":
			angles = [0]
			p_angles = [["x", 0]]
		elif usr_opt == "2":
			angles = random.choices([math.pi * i for i in [0, 0.5, 1, 1.5]], k=3)
			rot = [["x", 0], ["x", .5], ["x", 1], ["x", 1.5], ["z", 0.5], ["z", 1.5]]
			p_angles = random.choices([[i[0], math.pi * i[1]] for i in rot], k=4)
		elif usr_opt == "3":
			angles = [math.pi * i for i in [0, 0.5, 1, 1.5]]
			rot = [["x", 0], ["x", .5], ["x", 1], ["x", 1.5], ["z", 0.5], ["z", 1.5]]
			p_angles = [[i[0], math.pi * i[1]] for i in rot]

		while True:
			save = input("Save individual files? (y/n)")
			if save in ["y", "n"]: break
		rot_alg(xyz_1, xyz_2, angles, p_angles, geodes, save)
	start()




import os, math, functools
import numpy as np
from raachem.util.constants import elements, element_radii
from raachem.util.gen_purp import is_str_float

class XyzFile:
	def __init__(self,file_content):
		self.list = file_content
		if len(self.list) < 2: raise Exception(".xyz Object is empty?")
		elif not (str(self.list[1]).strip().isdigit() and len(str(self.list[1]).split()) == 1):
			print("{} is not a proper .xyz file\nAttempting to read it anyway!".format(self.list[0]))
			try_xyz = []
			for line in self.list:
				line = line.split()
				if len(line) != 4: continue
				if not all(is_str_float(line[i]) for i in range(1, 4)): continue
				if line[0] in elements[0:]:
					try_xyz.append(" ".join(line))
					continue
				try:
					line[0] = elements[int(line[0])]
					try_xyz.append(" ".join(line))
				except:
					raise Exception("Could not understand file {}".format(self.list[0]))
			try_xyz.insert(0,len(try_xyz))
			try_xyz.insert(1," ")
			try_xyz.insert(0,self.list[0])
			self.list = try_xyz
		self.list_l = [str(a).split() for a in self.list]
		#self.molecule.print_int_bond_map()
	def __add__(self,other):
		assert type(self) == type(other), "Operation '+' allowed only for two XYZ objects"
		new = [os.path.splitext(self.name())[0]+"_"+other.name(), str(self.n_atoms()+other.n_atoms()),
			   self.title()+" "+other.title(),*(self.form_cord_block() + other.form_cord_block())]
		return XyzFile(new)
	def __sub__(self, other):
		el_a = self.all_elements()
		el_b = other.all_elements()
		assert len(el_a) > len (el_b), "Can't subtract a larger structure from a smaller one"
		assert type(self) == type(other), "Operation '-' allowed only for two XYZ objects"
		idxs_to_rem = []
		for n in range(len(el_a) - len(el_b)):
			if all([True if el_a[n+i] == a else False for i,a in enumerate(el_b)]):
				idxs_to_rem = [c+n for c in range(len(el_b))]
				break
		if len(idxs_to_rem) ==  0: print("Could not subtract value!")
		xyz_cord = [a for idx,a in enumerate(self.form_cord_block()) if idx not in idxs_to_rem]
		new = [os.path.splitext(self.name())[0]+"-"+other.name(), str(self.n_atoms()-other.n_atoms()),
			   self.title()+"-"+other.title(),*xyz_cord]
		return XyzFile(new)
	def __str__(self):
		return "\n".join(self.return_print())
	@functools.lru_cache(maxsize=1)
	def name(self):
		if len(self.list[0]) == 0: raise Exception(".xyz Object has no name")
		return self.list[0]
	@functools.lru_cache(maxsize=1)
	def n_atoms(self):
		if any([len(str(self.list[1]).split()) != 1, not str(self.list[1]).isnumeric()]):
			raise Exception("First line of {} (.xyz type) file should contain only the number of atoms in the geometry!".format(self.name()))
		return int(self.list[1])
	@functools.lru_cache(maxsize=1)
	def title(self):
		return self.list[2]
	@functools.lru_cache(maxsize=1)
	def cord_block(self):
		cordinates = []
		for idx,line in enumerate(self.list_l):
			if idx <= 2: continue
			if idx >= self.n_atoms() + 3: continue
			if line[0] in elements:	cordinates.append(line)
			else: cordinates.append([elements[int(line[0])],*line[0:]])
		return cordinates
	@functools.lru_cache(maxsize=1)
	def form_cord_block(self):
		return ["{:<5}{:>20.6f}{:>20.6f}{:>20.6f}".format(x[0], *[float(x[a]) for a in [1, 2, 3]]) for x in self.cord_block()]
	@functools.lru_cache(maxsize=1)
	def cord_strip(self):
		return [line[1:] for line in self.cord_block()]
	@functools.lru_cache(maxsize=1)
	def all_elements(self):
		return [line[0] for line in self.cord_block()]
	@functools.lru_cache(maxsize=1)
	def elements(self):
		return list(dict.fromkeys(self.all_elements()))
	@functools.lru_cache(maxsize=1)
	def n_electrons(self):
		return sum(elements.index(e) for e in self.all_elements())
	@functools.lru_cache(maxsize=1)
	def return_print(self):
		return [str(self.n_atoms()),self.title(),*[l for l in self.form_cord_block()]]
	def print_file(self):
		print("======={}=======".format(self.name()))
		print("=======START=======")
		print("\n".join([l for l in self.return_print()]))
		print("========END========")
	def save_file(self,directory=None):
		if directory is None:
			file_path = os.path.splitext(os.path.join(os.getcwd(),self.name().replace(" ","")))[0]+".xyz"
		else:
			file_path = os.path.splitext(os.path.join(directory,self.name().replace(" ","")))[0]+".xyz"
		if os.path.exists(file_path):
			print("File {} already exists!".format(os.path.splitext(os.path.basename(file_path))[0] + ".xyz"))
			return
		with open(file_path,"w") as file:
			for line in self.return_print():file.write(str(line)+"\n")
		print("File {} saved!".format(os.path.splitext(os.path.basename(file_path))[0] + ".xyz"))
	def print_all(self):
		print("\n".join([l for l in self.list]))
	def displace(self,mult,displacement):
		cord_block = [[a,*[float(b[n])-c[n]*mult for n in range(3)]] for a,b,c in zip(self.all_elements(),self.cord_strip(),displacement)]
		cord_block = [" ".join([str(i) for i in l]) for l in cord_block]
		return XyzFile([self.name(),self.n_atoms(),self.title(),*cord_block])
	def rotate(self, angle, axis):
		"takes xyz object and returns xyz object rotated by angle over axis"
		assert axis in ("x", "y", "z"), "Only 'x','y' or 'z' axis are suported"
		if axis == "x":
			m_mat = [[1., 0., 0.], [0., math.cos(angle), -math.sin(angle)], [0., math.sin(angle), math.cos(angle)]]
		if axis == "y":
			m_mat = [[math.cos(angle), 0., math.sin(angle)], [0., 1., 0.], [-math.sin(angle), 0., math.cos(angle)]]
		if axis == "z":
			m_mat = [[math.cos(angle), -math.sin(angle), 0.], [math.sin(angle), math.cos(angle), 0.], [0., 0., 1.]]
		m_mat = np.array(m_mat, np.float64)
		rotated = np.array([i[1:4] for i in self.cord_block()], np.float64).transpose()
		rotated = np.matmul(m_mat,rotated).transpose()
		rotated = np.ndarray.tolist(rotated)
		rotated = [[b,*[str(n) for n in a]] for b,a in zip(self.all_elements(),rotated)]
		xyz_mat = [self.name(), self.n_atoms()," ",*[" ".join(a) for a in rotated]]
		return XyzFile(xyz_mat)
	def superimpose(self, other, num_atoms=0, print_step=False, ret = "geom",conv=18):
		"""Takes xyz object and returns xyz object rotated by angle over axis.
		Returns either the max_distance 'max_d' or final geometry 'geom' after rotations and superpositions"""
		def rotate(xyz,angle,axis):
			assert axis in ("x","y","z"), "Only 'x','y' or 'z' axis are suported"
			if axis == "x":
				m_mat = [[1., 0., 0.], [0., math.cos(angle), -math.sin(angle)], [0., math.sin(angle), math.cos(angle)]]
			if axis == "y":
				m_mat = [[math.cos(angle), 0., math.sin(angle)], [0., 1., 0.], [-math.sin(angle), 0., math.cos(angle)]]
			if axis == "z":
				m_mat = [[math.cos(angle), -math.sin(angle), 0.], [math.sin(angle), math.cos(angle), 0.], [0., 0., 1.]]
			m_mat = np.array(m_mat, np.float64)
			rotated = np.array(xyz, np.float64).transpose()
			rotated = np.matmul(m_mat,rotated).transpose()
			return np.ndarray.tolist(rotated)
		def calc_err(xyz_1, xyz_2, n_atms):
			n_atms = len(xyz_1) if n_atms == 0 else n_atms
			sq_dist = sum(sum(math.pow(c-d,2) for c,d in zip(a,b)) for a,b in zip(xyz_1[:n_atms],xyz_2))
			return math.sqrt(sq_dist / n_atms)
		def max_dist(xyz_a, xyz_b):
			return max(math.sqrt(sum(pow(c-d,2) for c,d in zip(a,b))) for a,b in zip(xyz_a,xyz_b))
		#----------------------
		last_error = None
		xyz_1 = [[float(a) for a in b] for b in other.std_cord(num_atoms).cord_strip()]
		xyz_2 = [[float(a) for a in b] for b in self.std_cord(num_atoms).cord_strip()]
		#Check atom correspondence
		for a,b,c in zip(range(len(self.all_elements()) if num_atoms == 0 else num_atoms),other.all_elements(),self.all_elements()):
			if b != c:
				atom_number = 'th' if 11<=a+1<=13 else {1:'st',2:'nd',3:'rd'}.get((a+1)%10, 'th')
				print("WARNING: {}{} atom pair doesn't not correspond to an element match: {} & {}".format(a+1,atom_number,b,c))
		if print_step: print("======ACTIONS======")
		#Start algorithm
		for num in range(conv):
			step_size = 1 / 2 ** num
			while True:
				rot = [[1, "x"], [1, "y"], [1, "z"], [-1, "x"], [-1, "y"], [-1, "z"]]
				movements = [rotate(xyz_2, step_size * i[0], i[1]) for i in rot]
				if ret == "max_d":
					last_error = max_dist(xyz_2, xyz_1)
					errors = [max_dist(i, xyz_1) for i in movements]
				else:
					last_error = calc_err(xyz_2, xyz_1, num_atoms)
					errors = [calc_err(i, xyz_1, num_atoms) for i in movements]
				best_m = errors.index(min(errors))
				if min(errors) < last_error:
					xyz_2 = movements[best_m]
					if print_step:
						msg = [step_size * rot[best_m][0], rot[best_m][1], calc_err(xyz_1, xyz_2, num_atoms)]
						print("Rotating {:.5f} radian in {}. RMSD = {:.5f}".format(*msg))
					continue
				else:
					if ret == "max_d" and max_dist(xyz_1, xyz_2) < 0.1:
						return True
					break
		if print_step: print("Final RMSD = {:.5f}".format(calc_err(xyz_1, xyz_2, num_atoms)))
		if print_step: print("========END========")
		if ret == "geom":
			cord_block = [" ".join([a,*[str(n) for n in b]]) for a,b in zip(self.all_elements(),xyz_2)]
			return XyzFile([self.name(),self.n_atoms(),self.title(),*cord_block])
		elif ret == "max_d":
			return False
	def std_cord(self, n_atoms=0):
		pure_cord = self.cord_strip() if n_atoms == 0 else self.cord_strip()[0:n_atoms]
		xyz_avg = [[float(n) for n in i] for i in pure_cord]
		xyz_avg = [sum([i[n] for i in xyz_avg]) / len(xyz_avg) for n in range(3)]
		xyz_avg = [[float(i[n]) - xyz_avg[n] for n in range(3)] for i in self.cord_strip()]
		xyz_avg = [[str(n) for n in a] for a in xyz_avg]
		xyz_avg = [" ".join([b,*a]) for b,a in zip(self.all_elements(),xyz_avg)]
		xyz_mat = [self.name(), self.n_atoms(), " ", *xyz_avg]
		return XyzFile(xyz_mat)
	def enantiomer(self):
		xyz = [" ".join([*a[0:-1],str(-float(a[-1]))]) for a in self.cord_block()]
		xyz_mat = [os.path.splitext(self.name())[0]+"_ent.xyz", self.n_atoms(), " ", *xyz]
		return XyzFile(xyz_mat)
	molecule = property(lambda self: Molecule(self.cord_block()))

class Molecule:
	def __init__(self,atom_list):
		assert type(atom_list) is list
		self.atom_list = [Atom(a,i) for i,a in enumerate(atom_list)]
		self.abc_angle(0,1,2)
		self.n_mol_ent()
	def __str__(self):
		return "\n".join([str(a) for a in self.atom_list])
	def int_bond_map(self):
		return [[b.int_bond_order(a) if a != b and b.int_bond_order(a) > 0.85 else None for a in self.atom_list] for b in self.atom_list]
	def ts_bond_map(self):
		return [[b.ts_bond_order(a) if a != b and b.int_bond_order(a) > 0.85 else None for a in self.atom_list] for b in self.atom_list]
	def print_int_bond_map(self):
		for a in self.atom_list:
			bonded = ["{:>3}{:>2}:{:.1f}".format(b.idx,b.element, b.int_bond_order(a)) for b in self.atom_list if a != b and b.int_bond_order(a) > 0.1]
			print("{:>3}{:>2}".format(a.idx,a.element),"-->",", ".join(bonded))
	def print_ts_bond_map(self):
		for a in self.atom_list:
			bonded = ["{:>3}{:>2}:{:.1f}".format(b.idx,b.element, b.ts_bond_order(a)) for b in self.atom_list if a != b and b.ts_bond_order(a) > 0.1]
			print("{:>3}{:>2}".format(a.idx,a.element),"-->",", ".join(bonded))
	def n_mol_ent(self, map=None):
		if map is None: map = self.int_bond_map()
		visited = [False for _ in map]
		n_entities = 0
		entities = []
		def check(idx,atoms=[]):
			visited[idx] = True
			atoms.append(idx)
			for ib, b in enumerate(map[idx]):
				if b is None: continue
				elif visited[ib]: continue
				else:
					print(f"Leaving {idx+1} to check on {ib+1} because of BO: {b}")
					check(ib, atoms)
			return atoms
		for ia,a in enumerate(map):
			if visited[ia]: continue
			else:
				print(f"Adding new entitie starting from {ia+1}")
				n_entities +=1
				entities.append(check(ia,[]))
		print("Visited\n",visited)
		print("n entitites\n", n_entities)
		print("entities\n", entities)

	def valid_idxs(func):
		def wrapper(obj,*list):
			assert all([type(n) is int for n in list]), "Atom indexes should be integers"
			assert all([n in range(len(obj.atom_list)) for n in list]), "Atom indexes are out of range"
			return func(obj,*list)
		return wrapper
	@valid_idxs
	def ab_distance(self,a,b):
		return self.atom_list[a].distance(self.atom_list[b])
	@valid_idxs
	def abc_angle(self,a,b,c):
		return self.atom_list[a].angle(self.atom_list[b],self.atom_list[c])
	@valid_idxs
	def abcd_dihedral(self,a,b,c,d):
		return self.atom_list[a].dihedral(self.atom_list[b],self.atom_list[c],self.atom_list[d])

class Atom:
	el_radii = dict(element_radii)
	def __init__(self,line,idx):
		assert type(line) is list
		assert len(line) == 4
		assert line[0] in elements
		assert all(is_str_float(a) for a in line[1:])
		self.idx = idx
		self.element = line[0]
		self.cord = [float(a) for a in line[1:]]
	def distance(self,other):
		return sum((b - a) ** 2 for a, b in zip(self.cord, other.cord)) ** 0.5
	def angle(self,other_a,other_b):
		a_a = np.array(self.cord)
		b_a = np.array(other_a.cord)
		c_a = np.array(other_b.cord)
		ba, bc = a_a - b_a, c_a - b_a
		cosine_angle = np.dot(ba, bc) / (np.linalg.norm(ba) * np.linalg.norm(bc))
		angle = np.arccos(cosine_angle)
		#print("Angle :",self.idx,other_a.idx,other_b.idx,"is:", "{:.2f}°".format(np.degrees(angle)))
		return angle
	def dihedral(self,other_a,other_b,other_c):
		p = np.array([self.cord,other_a.cord,other_b.cord,other_c.cord])
		# From: stackoverflow.com/questions/20305272/dihedral-torsion-angle-from-four-points-in-cartesian-coordinates-in-python
		b = p[:-1] - p[1:]
		b[0] *= -1
		v = np.array([np.cross(v, b[1]) for v in [b[0], b[2]]])
		# Normalize vectors
		v /= np.sqrt(np.einsum('...i,...i', v, v)).reshape(-1, 1)
		return np.degrees(np.arccos(v[0].dot(v[1])))
	def ts_bond_order(self,other):
		return math.exp((Atom.el_radii[self.element]/100 + Atom.el_radii[other.element]/100 - self.distance(other))/0.6)
	def int_bond_order(self,other):
		return math.exp((Atom.el_radii[self.element]/100 + Atom.el_radii[other.element]/100 - self.distance(other))/0.3)
	def __str__(self):
		return "{}{}".format(self.idx,self.element)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="raachem",
    version="0.0.8.post2",
    author="Ricardo Almir Angnes",
    author_email="ricardo_almir@hotmail.com",
    description="Tools I use for computational chemistry",
    long_description=long_description,
	license="MIT",
    url="https://github.com/ricalmang/raachem",
	keywords = ['chemistry'],
	install_requires = ["numpy","xlwt"],
    packages=setuptools.find_packages(),
	package_data = {"":["*.GAUSSIAN","*.ORCA","*.gbs"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


#!/usr/bin/env python
import os
directory=os.path.dirname(os.path.abspath(__file__))

def file_weeder(ext_to_weed):
	"""Looks up files with the extensions provided in current directory"""
	fulllist=os.listdir(os.path.dirname(os.path.abspath(__file__)))
	weeded_list=[]
	for extension in ext_to_weed:
		matching = [s for s in fulllist if extension in s]
		for match in matching:
			weeded_list.append(match)
	return sorted(weeded_list)

	
anr_script , files_for_p = [], []
all_files = file_weeder([".pbs"])
print("The following files will compose the script")
for idx,i in enumerate(all_files):
	if idx == 0:
		counter, a = 0, []
	counter += 1
	a.append(i)
	if counter == 3 or idx + 1 == len(all_files):
		files_for_p.append(a)
		counter , a = 0, [] 	
for line in files_for_p:
	print(" ".join(["{0}".format(a) for a in line]))

while True:
	n_items = input("How many of these items do you want to submit?\n 'a' for all\n'0' to cancel\n")
	if n_items == "a":
		n_items = int(len(all_files))
		break
	elif n_items == "0":
		print("Ok!, Exiting now!")
		exit()
	elif any([n_items == str(i+1) for i in range(len(all_files))]):
		n_items = int(n_items)
		break

for file in file_weeder([".pbs"])[:n_items]:
	anr_script.append("qsub {0}".format(file))
	anr_script.append("sleep 0.2")
	anr_script.append("rm {0}".format(file))
	anr_script.append("sleep 0.2")
anr_script.append("rm sub_s_name")

if anr_script:
	with open(os.path.join(directory,"sub_s_name"),"w+") as script:
		script.write("\n".join(anr_script))
os.system("chmod 755 sub_s_name")


